from logging import Logger, StreamHandler
import os
import re
from warnings import warn

from ansys.mapdl.core import __version__
from ansys.mapdl.core.commands import Commands
from ansys.mapdl.core.misc import is_float

# Because the APDL version has empty arguments, whereas the PyMAPDL
# doesn't have them. Hence the order of arguments is messed up.
COMMANDS_TO_NOT_BE_CONVERTED = {"INT1"}

FORMAT_OPTIONS = {
    "select": "W191,W291,W293,W391,E115,E117,E122,E124,E125,E225,E231,E301,E303,F401,F403",
    "max-line-length": 100,
}


def convert_script(
    filename_in,
    filename_out,
    loglevel="WARNING",
    auto_exit=True,
    line_ending=None,
    exec_file=None,
    macros_as_functions=True,
    use_function_names=True,
    show_log=False,
    add_imports=True,
    comment_solve=False,
    cleanup_output=True,
    header=True,
    print_com=True,
):
    """Converts an ANSYS input file to a python PyMAPDL script.

    Parameters
    ----------
    filename_in : str
        Filename of the ansys input file to read in.

    filename_out : str
        Filename of the python script to write a translation to.

    loglevel : str, optional
        Logging level of the ansys object within the script.

    auto_exit : bool, optional
        Adds a line to the end of the script to exit MAPDL.  Default
        ``True``.

    line_ending : str, optional
        When None, automatically determined by OS being used.

    macros_as_functions : bool, optional
        Attempt to convert MAPDL macros to python functions.

    use_function_names : bool, optional
        Convert MAPDL functions to ansys.mapdl.core.Mapdl class
        methods.  When ``True``, the MAPDL command "K" will be
        converted to ``mapdl.k``.  When ``False``, it will be
        converted to ``mapdl.run('k')``.

    show_log : bool, optional
        Print the converted commands using a logger (from ``logging``
        Python module).

    add_imports : bool, optional
        If ``True``, add the lines ``from ansys.mapdl.core import launch_mapdl``
        and ``mapdl = launch_mapdl(loglevel="WARNING")`` to the
        beginning of the output file. This option is useful if you
        are planning to use the output script from another mapdl
        session. See examples section.
        This option overrides ``auto_exit``.

    comment_solve : bool, optional
        If ``True``, it will pythonically comment the lines that
        contain ``"SOLVE"`` or ``"/EOF"``.

    cleanup_output : bool, optional
        If ``True`` the output is formatted using ``autopep8`` before writing
        the file or returning the string. This requires ``autopep8`` to be
        installed.

    header : bool, optional
        If ``True``, the default header is written in the first line
        of the output. If a string is provided, this string will be
        used as header.

    print_com : bool, optional
        Print command ``/COM`` arguments to python console.
        Defaults to ``True``.

    Returns
    -------
    list
        List of lines translated.

    Examples
    --------
    >>> from ansys.mapdl import core as pymapdl
    >>> from ansys.mapdl.core import examples
    >>> clines = pymapdl.convert_script(examples.vmfiles['vm1'], 'vm1.py')

    Converting a script and using it already in the same session.
    For this case, it is recommended to use
    :func:`convert_apdl_block() <ansys.mapdl.core.convert_apdl_block>`
    since you do not need to write the file.

    >>> from ansys.mapdl.core import launch_mapdl
    >>> from ansys.mapdl.core import examples
    >>> from ansys.mapdl.core import convert_script
    >>> in_file = examples.vmfiles['vm10']
    >>> filename = in_file.split('\\')[-1]
    >>> out_file = 'out_' + filename.replace('.dat', '.py')
    >>> output = convert_script(file, out_file, line_ending='\\n')
    >>> mapdl = launch_mapdl()
    >>> with open(out_file, 'r') as fid:
    ...     cmds = fid.read()
    >>> mapdl.input_strings(cmds.splitlines()[2:10])

    """
    with open(filename_in, "r") as fid:
        apdl_strings = fid.readlines()

    translator = _convert(
        apdl_strings=apdl_strings,
        loglevel=loglevel,
        auto_exit=auto_exit,
        line_ending=line_ending,
        exec_file=exec_file,
        macros_as_functions=macros_as_functions,
        use_function_names=use_function_names,
        show_log=show_log,
        add_imports=add_imports,
        comment_solve=comment_solve,
        cleanup_output=cleanup_output,
        header=header,
        print_com=print_com,
    )

    translator.save(filename_out)
    return translator.lines


def convert_apdl_block(
    apdl_strings,
    loglevel="WARNING",
    auto_exit=True,
    line_ending=None,
    exec_file=None,
    macros_as_functions=True,
    use_function_names=True,
    show_log=False,
    add_imports=True,
    comment_solve=False,
    cleanup_output=True,
    header=True,
    print_com=True,
):
    """Converts an ANSYS input string to a python PyMAPDL string.

    Parameters
    ----------
    apdl_string : str
        APDL strings or list of strings to convert.

    filename_out : str
        Filename of the python script to write a translation to.

    loglevel : str, optional
        Logging level of the ansys object within the script.

    auto_exit : bool, optional
        Adds a line to the end of the script to exit MAPDL.  Default
        ``True``.

    line_ending : str, optional
        When None, automatically determined by OS being used.

    macros_as_functions : bool, optional
        Attempt to convert MAPDL macros to python functions.

    use_function_names : bool, optional
        Convert MAPDL functions to ansys.mapdl.core.Mapdl class
        methods.  When ``True``, the MAPDL command "K" will be
        converted to ``mapdl.k``.  When ``False``, it will be
        converted to ``mapdl.run('k')``.

    show_log : bool, optional
        Print the converted commands using a logger (from ``logging``
        Python module).

    add_imports : bool, optional
        If ``True``, it add the next lines to the beginning of the
        output file:

        .. code:: python

           from ansys.mapdl.core import launch_mapdl
           mapdl = launch_mapdl(loglevel="WARNING")

        This option is useful if you are planning to use the output
        script from another mapdl session. See examples section.
        This option overrides ``'auto_exit'``.

    comment_solve : bool, optional
        If ``True``, pythonically comment the lines containing
        ``mapdl.solve`` or ``"/EOF"``.

    cleanup_output : bool, optional
        If ``True`` the output is formatted using ``autopep8`` before
        writing the file or returning the string.

    header : bool, optional
        If ``True``, the default header is written in the first line
        of the output. If a string is provided, this string will be
        used as header.
    print_com : bool, optional
        Print command ``/COM`` arguments to python console.
        Defaults to ``True``.

    Returns
    -------
    list
        List of lines translated.

    Examples
    --------
    Convert a script and use it in the same session.

    >>> from ansys.mapdl.core import examples, launch_mapdl, convert_apdl_block
    >>> in_file = examples.vmfiles['vm10']
    >>> filename = in_file.split('\\')[-1]
    >>> out_file = 'out_' + filename.replace('.dat', '.py')
    >>> cmds = convert_apdl_block(file, out_file, line_ending='\\n')
    >>> # Do any change in the text, for example:
    >>> cmds = cmds.replace('solve', '!solve')
    >>> mapdl = launch_mapdl()
    >>> mapdl.input_strings(cmds.splitlines()[2:10])

    """

    translator = _convert(
        apdl_strings,
        loglevel=loglevel,
        auto_exit=auto_exit,
        line_ending=line_ending,
        exec_file=exec_file,
        macros_as_functions=macros_as_functions,
        use_function_names=use_function_names,
        show_log=show_log,
        add_imports=add_imports,
        comment_solve=comment_solve,
        cleanup_output=cleanup_output,
        header=header,
        print_com=print_com,
    )

    if isinstance(apdl_strings, str):
        return translator.line_ending.join(translator.lines)
    return translator.lines


def _convert(
    apdl_strings,
    loglevel="WARNING",
    auto_exit=True,
    line_ending=None,
    exec_file=None,
    macros_as_functions=True,
    use_function_names=True,
    show_log=False,
    add_imports=True,
    comment_solve=False,
    cleanup_output=True,
    header=True,
    print_com=True,
):

    translator = FileTranslator(
        loglevel,
        line_ending,
        exec_file=exec_file,
        macros_as_functions=macros_as_functions,
        use_function_names=use_function_names,
        show_log=show_log,
        add_imports=add_imports,
        comment_solve=comment_solve,
        cleanup_output=cleanup_output,
        header=header,
        print_com=print_com,
    )

    if isinstance(apdl_strings, str):
        # os.linesep does not work well, so we are making sure
        # the line separation is appropriate.
        regx = f"[^\\r]({translator.line_ending})"
        if not re.search(regx, apdl_strings):
            if "\r\n" in apdl_strings:
                translator.line_ending = "\r\n"
            elif "\n" in apdl_strings:
                translator.line_ending = "\n"

        apdl_strings = apdl_strings.split(translator.line_ending)

    for line in apdl_strings:
        translator.translate_line(line)

    if auto_exit and add_imports:
        translator.write_exit()
    return translator


class Lines(list):
    def __init__(self, mute):
        self._log = Logger("convert_logger")
        self._setup_logger()
        self._mute = mute
        super().__init__()

    def append(self, item, mute=True):
        # append the item to itself (the list)
        if not self._mute:
            self._log.info(msg=f"Converted: '{item}'")
        super(Lines, self).append(item)

    def _setup_logger(self):
        stdhdl = StreamHandler()
        stdhdl.setLevel(10)
        stdhdl.set_name("stdout")
        self._log.addHandler(stdhdl)
        self._log.propagate = True


class FileTranslator:
    obj_name = "mapdl"
    indent = ""
    non_interactive = False

    def __init__(
        self,
        loglevel="INFO",
        line_ending=None,
        exec_file=None,
        macros_as_functions=True,
        use_function_names=True,
        show_log=False,
        add_imports=True,
        comment_solve=False,
        cleanup_output=True,
        header=True,
        print_com=True,
    ):
        self._non_interactive_level = 0
        self.lines = Lines(mute=not show_log)
        self._functions = []
        if line_ending:
            self.line_ending = line_ending
        else:
            self.line_ending = os.linesep
        self.macros_as_functions = macros_as_functions
        self._infunction = False
        self.use_function_names = use_function_names
        self.comment = ""
        self._add_imports = add_imports
        self._comment_solve = comment_solve
        self.cleanup_output = cleanup_output
        self._header = header
        self.print_com = print_com

        self.write_header()
        if self._add_imports:
            self.initialize_mapdl_object(loglevel, exec_file)

        self._valid_commands = dir(Commands)
        self._block_commands = {
            "NBLO": "NBLOCK",
            "EBLO": "EBLOCK",
            "BFBL": "BFBLOCK",
            "BFEB": "BFEBLOCK",
            "PREA": "PREAD",
            "SFEB": "SFEBLOCK",
        }  # Way out: '-1' , 'END PREAD'

        self._enum_block_commands = {
            "CMBL": "CMBLOCK",
        }  # Commands where you need to count the number of lines.

        _NON_INTERACTIVE_COMMANDS = {
            "*CRE": "*CREATE",
            "*VWR": "*VWRITE",
            "*VRE": "*VREAD",
        }

        self._non_interactive_commands = (
            list(_NON_INTERACTIVE_COMMANDS)
            + list(self._block_commands)
            + list(self._enum_block_commands)
        )

        self._block_count = 0
        self._block_count_target = 0
        self._in_block = False
        self._block_current_cmd = None

    def write_header(self):
        if isinstance(self._header, bool):
            if self._header:
                header = (
                    f'"""Script generated by ansys-mapdl-core version {__version__}"""'
                )
                self.lines.append(header)
        elif isinstance(self._header, str):
            self.lines.append(f'"""{self._header}"""')

        else:
            raise TypeError(
                "The keyword argument 'header' should be a string or a boolean."
            )

    def write_exit(self):
        self.lines.append(f"{self.obj_name}.exit()")

    def format_using_autopep8(self, text=None):
        """Format internal `self.lines` with autopep8.

        Parameters
        ----------
        text : str, optional
            Text to format instead of `self.lines`. For development use.

        """
        if self.cleanup_output:

            try:
                import autopep8
            except ModuleNotFoundError:  # pragma: no cover
                warn(
                    "Install `autopep8` to use this feature with\n"
                    "`pip install autopep8`"
                )
                return

            if not text:
                text = self.line_ending.join(self.lines)
                self.lines = autopep8.fix_code(text).splitlines()

            else:  # pragma: no cover
                # for development purposes
                return autopep8.fix_code(text)

    def save(self, filename, format_autopep8=True):
        """Saves lines to file"""
        if os.path.isfile(filename):
            os.remove(filename)

        # Making sure we write python string with double slash.
        # We are not expecting other type of unicode symbols.
        self.lines = [each_line.replace("\\", "\\\\") for each_line in self.lines]

        # Try to format the file using AutoPEP8
        self.format_using_autopep8()

        with open(filename, "w") as f:
            f.write(self.line_ending.join(self.lines))

    def initialize_mapdl_object(self, loglevel, exec_file):
        """Initializes ansys object as lines"""
        core_module = "ansys.mapdl.core"  # shouldn't change
        self.lines.append(f"from {core_module} import launch_mapdl")

        if exec_file:
            exec_file_parameter = f'"{exec_file}", '
        else:
            exec_file_parameter = ""

        if self.print_com:
            line = f'{self.obj_name} = launch_mapdl({exec_file_parameter}loglevel="{loglevel}", print_com=True)'
        else:
            line = f'{self.obj_name} = launch_mapdl({exec_file_parameter}loglevel="{loglevel}")'
        self.lines.append(line)

    @property
    def line_ending(self):
        return self._line_ending

    @line_ending.setter
    def line_ending(self, line_ending):
        if line_ending not in ["\n", "\r\n"]:
            raise ValueError('Line ending must be either "\\n", "\\r\\n"')
        self._line_ending = line_ending

    def translate_line(self, line):
        """Converts a single line from an ANSYS APDL script"""
        self.comment = ""
        original_line = line.replace("\r\n", "").replace(
            "\n", ""
        )  # It is needed for the nblock, eblock since they have spaces before the numbers
        line = line.strip()
        line = line.replace('"', "'")

        if self._in_block:
            self._block_count += 1

        if (
            self._in_block
            and self._block_count >= self._block_count_target
            and self._block_count_target
        ):
            self._in_block = False
            self.end_non_interactive()
            self._block_count = 0
            self._block_current_cmd = None

        # check if line contains a comment
        if "!" in line:
            if "'!'" in line or '"!"' in line:
                pass
            elif line[0] == "!":  # entire line is a comment
                self.comment = line.replace("!", "").strip()
                self.store_comment()
                return
            else:  # command and in-line comment
                split_line = line.split("!")
                line = split_line[0]
                self.comment = " ".join(split_line[1:])
                self.comment = self.comment.lstrip()

        if not line:
            return

        # Cleaning ending empty arguments.
        # Because of an extra comma added to toffst command when generating ds.dat.
        line_ = line.split(",")[::-1]  # inverting order
        for ind, each in enumerate(line_):
            if each:
                break
            else:
                line_.pop(ind)
        line = ",".join(line_[::-1])

        # remove trailing comma
        line = line[:-1] if line[-1] == "," else line

        cmd_ = line.split(",")[0].upper()

        if cmd_[:4] in ["SOLV", "LSSO"] and self._comment_solve:
            self.store_command(
                "com", ["The following line has been commented due to `comment_solve`:"]
            )
            self.store_command("com", [line])
            return

        if cmd_[:4] == "/COM":
            # It is a comment
            self.store_command("com", [line[5:]])
            return

        if cmd_ == "*DO":
            self.start_non_interactive()
            self.store_run_command(line)
            return

        if cmd_ in ["*ENDDO", "*ENDIF"]:
            self.store_run_command(line)
            self.end_non_interactive()
            return

        if self.output_to_file(line):
            self.start_non_interactive()
            self.store_run_command(line)
            return

        if self.output_to_default(line):
            self.store_run_command(line)
            self.store_run_command("/GOPR")  # Adding gopr to ensure printing out
            self.end_non_interactive()
            return

        if cmd_ == "/VERIFY":
            self.store_run_command("FINISH")
            self.store_run_command(line)
            self.store_run_command("/PREP7")
            return

        if line[:4].upper() == "*REP":
            if not self.non_interactive:
                prev_cmd = self.lines.pop(-1)
                self.start_non_interactive()
                new_prev_cmd = (
                    "    " + prev_cmd
                )  # Since we are writing in self.lines we need to add the indentation by ourselves.
                self.lines.append(new_prev_cmd)
                self.store_run_command(
                    line
                )  # Using run but it could be `store_command`
                self.end_non_interactive()
                return

        if line[:4].upper() in COMMANDS_TO_NOT_BE_CONVERTED:
            self.store_run_command(line)
            return

        if line[:4].upper() == "/TIT":  # /TITLE
            parameters = line.split(",")[1:]
            return self.store_command("title", ["".join(parameters).strip()])

        if line[:4].upper() == "*GET":
            if self.non_interactive:  # gives error
                self.store_run_command(line)
                return
            else:
                parameters = line.split(",")[1:]
                return self.store_command("get", parameters)

        if line[:4].upper() == "/NOP":
            self.comment = (
                "It is not recommended to use '/NOPR' in a normal PyMAPDL session."
            )
            self.store_under_scored_run_command(line)
            return

        if line[:6].upper() == "/PREP7":
            return self.store_command("prep7", [])

        if "*END" in line:
            if self.macros_as_functions:
                self.store_empty_line()
                self.store_empty_line()
                self.indent = self.indent[4:]
                self._infunction = False
                if not self._in_block:
                    self.end_non_interactive()
                return
            else:
                self.store_run_command(line)
                if not self._in_block:
                    self.end_non_interactive()
                return

        # check for if statement
        if line[:3].upper() == "*IF" or "*IF" in line.upper():
            self.start_non_interactive()
            self.store_run_command(line)
            return

        # check if line ends non-interactive
        if line[0] == "(":
            if not self.non_interactive:
                warn(
                    "\nPossible invalid line:\n%s\n" % line
                    + "This line requires a *VWRITE beforehand."
                    + "The previous line is: \n%s\n\n" % self.lines[-1]
                )
            self.store_run_command(line)
            if (
                not self._in_block
            ):  # To escape cmds that require (XX) but they are not in block
                self.end_non_interactive()
            return
        elif line[:4] == "*USE" and self.macros_as_functions:
            items = line.split(",")
            func_name = items[1].strip()
            if func_name in self._functions:
                args = ", ".join(items[2:])
                self.lines.append(f"{func_name}({args})")
                return

        # check if a line is setting a variable
        items = line.split(",")
        if "=" in items[0]:  # line sets a variable:
            self.store_run_command(line)
            return

        command = items[0].lower().strip()
        parameters = items[1:]
        if not command:
            self.store_empty_line()
            return

        if line == "-1" or line == "END PREAD":  # End of block commands
            self.store_run_command(line)
            self._in_block = False
            self.end_non_interactive()
            return

        # check valid command
        if command not in self._valid_commands:
            cmd = line[:4].upper()
            if cmd == "*CRE":  # creating a function
                if self.macros_as_functions:
                    self.start_function(items[1].strip())
                    return
                else:
                    self.start_non_interactive()

            elif cmd in self._non_interactive_commands:
                if cmd in self._block_commands:
                    self._in_block = True
                    self._block_count = 0
                    self._block_count_target = 0

                elif cmd in self._enum_block_commands:
                    self._in_block = True
                    self._block_count = 0
                    if cmd == "CMBL":  # In cmblock
                        # CMBLOCK,Cname,Entity,NUMITEMS,,,,,KOPT
                        numitems = int(line.split(",")[3])
                        _block_count_target = (
                            numitems // 8 + 1 if numitems % 8 != 0 else numitems // 8
                        )
                        self._block_count_target = (
                            _block_count_target + 2
                        )  # because the cmd line and option line.

                self._block_current_cmd = cmd
                self.start_non_interactive()

            if self._in_block and cmd not in self._non_interactive_commands:
                self.store_run_command(original_line)
            else:
                self.store_run_command(line)

        elif self.use_function_names:
            self.store_command(command, parameters)
        else:
            self.store_run_command(line)

    def start_function(self, func_name):
        self._functions.append(func_name)
        self.store_empty_line()
        self.store_empty_line()
        self._infunction = True
        spacing = " " * (len(func_name) + 5)
        line = "def %s(%s," % (
            func_name,
            ", ".join(["ARG%d=''" % i for i in range(1, 7)]),
        )
        line += "%s%s," % (spacing, ", ".join(["ARG%d=''" % i for i in range(7, 13)]))
        line += "%s%s):" % (spacing, ", ".join(["ARG%d=''" % i for i in range(13, 19)]))
        self.lines.append(line)
        self.indent = self.indent + "    "

    def store_under_scored_run_command(self, command):
        self.store_run_command(command, run_underscored=True)

    def store_run_command(self, command, run_underscored=False):
        """Stores pyansys.ANSYS command that cannot be broken down
        into a function and parameters.
        """
        if run_underscored:
            underscore = "_"
        else:
            underscore = ""

        if self._infunction and "ARG" in command:
            args = []
            for i in range(1, 19):
                arg = "ARG%d" % i
                c = 0
                if arg in command:
                    command = command.replace(arg, "{%d:s}" % c)
                    args.append(arg)
                    c += 1

            line = '%s%s.%srun("%s".format(%s))' % (
                self.indent,
                self.obj_name,
                underscore,
                command,
                ", ".join(args),
            )

        elif self.comment:
            line = '%s%s.%srun("%s")  # %s' % (
                self.indent,
                self.obj_name,
                underscore,
                command,
                self.comment,
            )
        else:
            line = '%s%s.%srun("%s")' % (
                self.indent,
                self.obj_name,
                underscore,
                command,
            )
        self.lines.append(line)

    def store_comment(self):
        """Stores a line containing only a comment"""
        line = f"{self.indent}# {self.comment}"
        self.lines.append(line)

    def store_empty_line(self):
        """Stores an empty line"""
        self.lines.append("")

    def store_command(self, function, parameters):
        """Stores a valid pyansys function with parameters"""
        parsed_parameters = []
        for parameter in parameters:
            parameter = parameter.strip()
            if is_float(parameter):
                parsed_parameters.append(parameter)
            elif "ARG" in parameter and self._infunction:
                parsed_parameters.append("%s" % parameter)
            else:
                # Removing strings '' and "" because they are going to be added by the converter module.
                if parameter.startswith("'") and parameter.endswith("'"):
                    parameter = parameter[1:-1]
                if parameter.startswith('"') and parameter.endswith('"'):
                    parameter = parameter[1:-1]
                parsed_parameters.append(f'"{parameter}"')

        parameter_str = ", ".join(parsed_parameters)
        if self.comment:
            line = "%s%s.%s(%s)  # %s" % (
                self.indent,
                self.obj_name,
                function,
                parameter_str,
                self.comment,
            )
        else:
            line = "%s%s.%s(%s)" % (self.indent, self.obj_name, function, parameter_str)

        self.lines.append(line)

    def start_non_interactive(self):
        self._non_interactive_level += 1
        if self.non_interactive:
            return
        line = f"{self.indent}with {self.obj_name}.non_interactive:"
        self.lines.append(line)
        self.non_interactive = True
        self.indent = self.indent + "    "

    def end_non_interactive(self):
        self._non_interactive_level -= 1
        if self._non_interactive_level == 0:
            self.non_interactive = False
            self.indent = self.indent[4:]

    def output_to_file(self, line):
        """Return if an APDL line is redirecting to a file."""
        if line[:4].upper() == "/OUT":
            # We are redirecting the output to somewhere, probably a file.
            # Because of the problem with the ansys output, we need to execute
            # this in non_interactive mode.
            output_cmd = line.strip().upper().split(",")
            if len(output_cmd) > 1:
                opt1 = output_cmd[1].strip().upper()
                if opt1 != "TERM":
                    # A file is supplied.
                    return True

        if (
            line[:4].upper() == "*CFO"
        ):  # any([each[0:4] in '*CFOPEN' for each in dir(Commands)])
            # We might not need going into interactive mode for *CFOPEN/*CFCLOSE
            return True

        return False

    def output_to_default(self, line):
        if line[:4].upper() == "/OUT":
            # We are redirecting the output to somewhere, probably a file.
            # Because of the problem with the ansys output, we need to execute
            # this in non_interactive mode.
            output_cmd = line.strip().upper().split(",")
            if len(output_cmd) == 1:
                return True
            elif len(output_cmd) > 1:
                opt1 = output_cmd[1].strip().upper()
                if opt1 == "TERM":
                    # A file is supplied.
                    return True
        if line[:4].upper() in "*CFCLOSE":
            # We might not need going into interactive mode for *CFOPEN/*CFCLOSE
            return True

        return False



"""Store parameters for a PyMAPDL-specific theme for pyvista"""
from ansys.mapdl.core import _HAS_PYVISTA

if _HAS_PYVISTA:
    from pyvista import themes

    base_class = themes.DefaultTheme

else:  # pragma: no cover

    class myEmptyClass:
        pass

    base_class = myEmptyClass


class MapdlTheme(base_class):
    """PyMAPDL-specific theme for pyvista.

    Theme includes the following defaults:

    - ``'jet'`` (rainbow) colormap
    - Interactive plots
    - ``'Courier'`` for the font family
    - ``'PyMAPDL'`` as the plot title

    Examples
    --------
    Create a custom theme with unique parameters from the base MapdlTheme.

    >>> from ansys.mapdl import core as pymapdl
    >>> my_theme = pymapdl.MapdlTheme()
    >>> my_theme.background = 'white'
    >>> my_theme.cmap = 'jet'  # colormap
    >>> my_theme.axes.show = False
    >>> my_theme.show_scalar_bar = False

    Apply this theme to element plotting.

    >>> mapdl.eplot(theme=theme)

    Apply this theme to area plotting.

    >>> mapdl.aplot(theme=theme)

    """

    def __init__(self):
        """Initialize the theme."""
        super().__init__()
        self.cmap = "jet"
        self.interactive = True
        self.font.family = "courier"
        self.title = "PyMAPDL"



from .core import _QueryExecution


class _ComponentQueries(_QueryExecution):
    _mapdl = None

    def centrx(self, e: int) -> float:
        """Return the x coordinate of the element centroid.

        Fetches centroid X-coordinate of element ``e`` in global
        Cartesian coordinate system. Centroid is determined from the
        selected nodes on the element.

        Parameters
        ----------
        e : int
            The element number of the element to be considered.

        Returns
        -------
        float
            The centroid coordinate.

        Examples
        --------
        Here we construct a line between the coordinates ``(0, 0, 0)``
        and ``(1, 2, 3)`` then find the centroid x-coordinate of this
        element.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> n0 = mapdl.n(1, 0, 0, 0)
        >>> n1 = mapdl.n(2, 1, 2, 3)
        >>> mapdl.et(1, 'LINK11')
        >>> e0 = mapdl.e(n0, n1)
        >>> mapdl.queries.centrx(e0)
        0.5
        """
        return self._run_query(f"CENTRX({e})", integer=False)

    def centry(self, e: int) -> float:
        """Return the y coordinate of the element centroid.

        Fetches centroid Y-coordinate of element ``e`` in global
        Cartesian coordinate system. Centroid is determined from the
        selected nodes on the element.

        Parameters
        ----------
        e : int
            The element number of the element to be considered.

        Returns
        -------
        float
            The centroid coordinate.

        Examples
        --------
        Here we construct a line between the coordinates (0, 0, 0) and
        (1, 2, 3) then find the centroid y-coordinate of this element.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> n0 = mapdl.n(1, 0, 0, 0)
        >>> n1 = mapdl.n(2, 1, 2, 3)
        >>> mapdl.et(1, 'LINK11')
        >>> e0 = mapdl.e(n0, n1)
        >>> mapdl.queries.centry(e0)
        1.0
        """
        return self._run_query(f"CENTRY({e})", integer=False)

    def centrz(self, e: int) -> float:
        """Return the z coordinate of the element centroid.

        Fetches centroid Z-coordinate of element ``e`` in global
        Cartesian coordinate system. Centroid is determined from the
        selected nodes on the element.

        Parameters
        ----------
        e : int
            The element number of the element to be considered.

        Returns
        -------
        float
            The centroid coordinate.

        Examples
        --------
        Here we construct a line between the coordinates (0, 0, 0) and
        (1, 2, 3) then find the centroid z-coordinate of this element.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> n0 = mapdl.n(1, 0, 0, 0)
        >>> n1 = mapdl.n(2, 1, 2, 3)
        >>> mapdl.et(1, 'LINK11')
        >>> e0 = mapdl.e(n0, n1)
        >>> mapdl.queries.centrz(e0)
        1.5
        """
        return self._run_query(f"CENTRZ({e})", integer=False)

    def nx(self, n: int) -> float:
        """Return the x coordinate of a node.

        Fetches X-coordinate of node ``n`` in the active coordinate
        system.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
             Coordinate of node

        Examples
        --------
        Here we construct a simple box and mesh it with elements. Then
        we use the ``nx`` method from the ``queries`` attribute to
        find the x-coordinate of the 10th node.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SOLID5')
        >>> mapdl.block(0, 10, 0, 20, 0, 30)
        >>> mapdl.esize(2)
        >>> mapdl.vmesh('ALL')
        >>> mapdl.queries.nx(10)
        0.0
        """
        return self._run_query(f"NX({n})", integer=False)

    def ny(self, n: int) -> float:
        """Return the y coordinate of a node.

        Fetches Y-coordinate of node ``n`` in the active coordinate
        system.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
             Coordinate of node

        Examples
        --------
        Here we construct a simple box and mesh it with elements. Then
        we use the the ``ny`` method from the ``queries`` attribute to
        find the y-coordinate of the 10th node.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SOLID5')
        >>> mapdl.block(0, 10, 0, 20, 0, 30)
        >>> mapdl.esize(2)
        >>> mapdl.vmesh('ALL')
        >>> mapdl.queries.ny(10)
        4.0
        """
        return self._run_query(f"NY({n})", integer=False)

    def nz(self, n: int) -> float:
        """Return the z coordinate of a node.

        Fetches Z-coordinate of node ``n`` in the active coordinate
        system.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
             Coordinate of node

        Examples
        --------
        Here we construct a simple box and mesh it with elements. Then
        we use the ``nz`` attribute from the ``queries`` attribute to
        find the z-coordinate of the 10th node.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SOLID5')
        >>> mapdl.block(0, 10, 0, 20, 0, 30)
        >>> mapdl.esize(2)
        >>> mapdl.vmesh('ALL')
        >>> mapdl.queries.nz(10)
        0.0
        """
        return self._run_query(f"NZ({n})", integer=False)

    def kx(self, k: int) -> float:
        """Return the x coordinate of a keypont.

        X-coordinate of keypoint ``k`` in the active coordinate system.

        Parameters
        ----------
        k : int
            Keypoint number to be considered.

        Returns
        -------
        float
            Coordinate of the keypoint.

        Examples
        --------
        Here we add a single keypoint, then use ``kx`` to extract the
        x-coordinate of it.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.k(1, 0, 1, 2)
        >>> mapdl.queries.kx(1)
        0.0
        """
        return self._run_query(f"KX({k})", integer=False)

    def ky(self, k: int) -> float:
        """Return the y coordinate of a keypont.

        Y-coordinate of keypoint ``k`` in the active coordinate system.

        Parameters
        ----------
        k : int
            Keypoint number to be considered.

        Returns
        -------
        float
            Coordinate of the keypoint.

        Examples
        --------
        Here we add a single keypoint, then use ``ky`` to extract the
        y-coordinate of it.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.k(1, 0, 1, 2)
        >>> mapdl.queries.ky(1)
        1.0
        """
        return self._run_query(f"KY({k})", integer=False)

    def kz(self, k: int) -> float:
        """Return the z coordinate of a keypont.

        Z-coordinate of keypoint ``k`` in the active coordinate system.

        Parameters
        ----------
        k : int
            Keypoint number to be considered.

        Returns
        -------
        float
            Coordinate of the keypoint.

        Examples
        --------
        Here we add a single keypoint, then use ``kz`` to extract the
        z-coordinate of it.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.k(1, 0, 1, 2)
        >>> mapdl.queries.kz(1)
        2.0
        """
        return self._run_query(f"KZ({k})", integer=False)


class _InverseGetComponentQueries(_QueryExecution):
    _mapdl = None

    def node(self, x: float, y: float, z: float) -> int:
        """Return node closest to coordinate ``(x, y, z)``.

        Number of the selected node nearest the `x`, `y`, `z` point.
        (In the active coordinate system, lowest number for coincident
        nodes.) A number higher than the highest node number indicates
        that the node is internal (generated by program).

        Parameters
        ----------
        x : float
            X-coordinate in the active coordinate system
        y : float
            Y-coordinate in the active coordinate system
        z : float
            Z-coordinate in the active coordinate system

        Returns
        -------
        int
            Node number

        Examples
        --------
        In this example we construct a solid cube and mesh it. Then we
        use ``queries.node`` to find the node closest to the centre of
        the cube. Using this we can extract the coordinates of this
        node and see how close to the centre the node is.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SOLID5')
        >>> mapdl.block(0, 10, 0, 10, 0, 10)
        >>> mapdl.esize(3)
        >>> mapdl.vmesh('ALL')
        >>> node_number = mapdl.queries.node(5., 5., 5.)
        >>> node_number
        112
        >>> x = mapdl.queries.nx(node_number)
        >>> y = mapdl.queries.ny(node_number)
        >>> z = mapdl.queries.nz(node_number)
        >>> (x, y, z)
        (5.0, 5.0, 5.0)
        """
        return self._run_query(f"NODE({x},{y},{z})", integer=True)

    def kp(self, x: float, y: float, z: float) -> int:
        """Return keypoint closest to coordinate ``(x, y, z)``.

        Number of the selected keypoint nearest the `x`, `y`, `z` point.

        In the active coordinate system, lowest number for coincident
        keypoints.

        Parameters
        ----------
        x : float
            X-coordinate in the active coordinate system
        y : float
            Y-coordinate in the active coordinate system
        z : float
            Z-coordinate in the active coordinate system

        Returns
        -------
        int
            Keypoint number

        Examples
        --------
        In this example we construct a simple triangle of keypoints in
        3D and then find the keypoint closest to the point (1, 1, 1).

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.k(1, 0, 1, 2)
        >>> mapdl.k(2, 1, 2, 0)
        >>> mapdl.k(3, 2, 0, 1)
        >>> mapdl.queries.kp(1., 1., 1.)
        1
        """
        return self._run_query(f"KP({x},{y},{z})", integer=True)


class _DisplacementComponentQueries(_QueryExecution):
    _mapdl = None

    def rotx(self, n: int) -> float:
        """Returns x-component of rotational displacement at a node.

        X-component of rotational displacement at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Rotational displacement of the node.

        Examples
        --------
        This example has been adapted from the example script
        :ref:`ref_rotational_displacement_example`. We create a square
        of shell material, apply a displacement perpendicular to the
        plane of the material, and then solve.

        Then we can use ``queries.rotx`` to get the x-component rotational
        displacement at the middle node on the deformed edge.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SHELL181')
        >>> mapdl.mp("EX", 1, 2e5)
        >>> mapdl.rectng(0, 1, 0, 1)
        >>> mapdl.sectype(1, "SHELL")
        >>> mapdl.secdata(0.1)
        >>> mapdl.esize(0.2)
        >>> mapdl.amesh("all")
        >>> mapdl.run('/SOLU')
        >>> mapdl.antype('STATIC')
        >>> mapdl.nsel("s", "loc", "x", 0)
        >>> mapdl.d("all", "all")
        >>> mapdl.nsel("s", "loc", "x", 1)
        >>> mapdl.d("all", "uz", -0.1)
        >>> mapdl.allsel("all")
        >>> mapdl.solve()
        >>> node = mapdl.queries.node(1, 0.5, 0)
        >>> mapdl.queries.rotx(node)
        -0.0002149851187
        """
        return self._run_query(f"ROTX({n})", integer=False)

    def roty(self, n: int) -> float:
        """Returns y-component of rotational displacement at a node.

        Y-component of rotational displacement at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Rotational displacement of the node.

        Examples
        --------
        This example has been adapted from the example script
        :ref:`ref_rotational_displacement_example`. We create a square
        of shell material, apply a displacement perpendicular to the
        plane of the material, and then solve.

        Then we can use ``queries.roty`` to get the y-component rotational
        displacement at the middle node on the deformed edge.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SHELL181')
        >>> mapdl.mp("EX", 1, 2e5)
        >>> mapdl.rectng(0, 1, 0, 1)
        >>> mapdl.sectype(1, "SHELL")
        >>> mapdl.secdata(0.1)
        >>> mapdl.esize(0.2)
        >>> mapdl.amesh("all")
        >>> mapdl.run('/SOLU')
        >>> mapdl.antype('STATIC')
        >>> mapdl.nsel("s", "loc", "x", 0)
        >>> mapdl.d("all", "all")
        >>> mapdl.nsel("s", "loc", "x", 1)
        >>> mapdl.d("all", "uz", -0.1)
        >>> mapdl.allsel("all")
        >>> mapdl.solve()
        >>> node = mapdl.queries.node(1, 0.5, 0)
        >>> mapdl.queries.roty(node)
        0.1489593933
        """
        return self._run_query(f"ROTY({n})", integer=False)

    def rotz(self, n: int) -> float:
        """Returns z-component of rotational displacement at a node.

        Z-component of rotational displacement at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Rotational displacement of the node.

        Examples
        --------
        This example has been adapted from the example script
        :ref:`ref_rotational_displacement_example`. We create a square
        of shell material, apply a displacement perpendicular to the
        plane of the material, and then solve.

        Then we can use ``queries.rotz`` to get the z-component rotational
        displacement at the middle node on the deformed edge.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SHELL181')
        >>> mapdl.mp("EX", 1, 2e5)
        >>> mapdl.rectng(0, 1, 0, 1)
        >>> mapdl.sectype(1, "SHELL")
        >>> mapdl.secdata(0.1)
        >>> mapdl.esize(0.2)
        >>> mapdl.amesh("all")
        >>> mapdl.run('/SOLU')
        >>> mapdl.antype('STATIC')
        >>> mapdl.nsel("s", "loc", "x", 0)
        >>> mapdl.d("all", "all")
        >>> mapdl.nsel("s", "loc", "x", 1)
        >>> mapdl.d("all", "uz", -0.1)
        >>> mapdl.allsel("all")
        >>> mapdl.solve()
        >>> node = mapdl.queries.node(1, 0.5, 0)
        >>> mapdl.queries.rotz(node)
        0.0
        """
        return self._run_query(f"ROTZ({n})", integer=False)

    def ux(self, n: int) -> float:
        """Returns x-component of structural displacement at a node.

        X-component of structural displacement at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Displacement of node

        Examples
        --------
        In this example we create a simple block of 6 cubic elements,
        fix one end in place, and then bend the other perpendicular to
        it. We can then examine the displacement of one of the nodes
        in the x-direction at the deformed end (node number 7).

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SOLID5')
        >>> mapdl.block(0, 10, 0, 20, 0, 30)
        >>> mapdl.esize(10)
        >>> mapdl.vmesh('ALL')
        >>> mapdl.mp('EX', 1, 210E9)
        >>> mapdl.nsel('S', 'LOC', 'Z', 0)
        >>> mapdl.d('ALL', 'UX')
        >>> mapdl.d('ALL', 'UY')
        >>> mapdl.d('ALL', 'UZ')
        >>> mapdl.nsel('S', 'LOC', 'Z', 30)
        >>> mapdl.f('ALL', 'FY', 1000)
        >>> mapdl.run('/SOLU')
        >>> mapdl.antype('STATIC')
        >>> mapdl.solve()
        >>> mapdl.finish()
        >>> mapdl.queries.ux(7)
        1.549155634e-07

        """
        return self._run_query(f"UX({n})", integer=False)

    def uy(self, n: int) -> float:
        """Returns y-component of structural displacement at a node.

        Y-component of structural displacement at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Displacement of node

        Examples
        --------
        In this example we create a simple block of 6 cubic elements,
        fix one end in place, and then bend the other perpendicular to
        it. We can then examine the displacement of one of the nodes
        in the y-direction at the deformed end (node number 7).

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SOLID5')
        >>> mapdl.block(0, 10, 0, 20, 0, 30)
        >>> mapdl.esize(10)
        >>> mapdl.vmesh('ALL')
        >>> mapdl.mp('EX', 1, 210E9)
        >>> mapdl.nsel('S', 'LOC', 'Z', 0)
        >>> mapdl.d('ALL', 'UX')
        >>> mapdl.d('ALL', 'UY')
        >>> mapdl.d('ALL', 'UZ')
        >>> mapdl.nsel('S', 'LOC', 'Z', 30)
        >>> mapdl.f('ALL', 'FY', 1000)
        >>> mapdl.run('/SOLU')
        >>> mapdl.antype('STATIC')
        >>> mapdl.solve()
        >>> mapdl.finish()
        >>> mapdl.queries.uy(7)
        5.803680779e-10

        """
        return self._run_query(f"UY({n})", integer=False)

    def uz(self, n: int) -> float:
        """Returns z-component of structural displacement at a node.

        Z-component of structural displacement at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Displacement of node

        Examples
        --------
        In this example we create a simple block of 6 cubic elements,
        fix one end in place, and then bend the other perpendicular to
        it. We can then examine the displacement of one of the nodes
        in the z-direction at the deformed end (node number 7).

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> mapdl.et(1, 'SOLID5')
        >>> mapdl.block(0, 10, 0, 20, 0, 30)
        >>> mapdl.esize(10)
        >>> mapdl.vmesh('ALL')
        >>> mapdl.mp('EX', 1, 210E9)
        >>> mapdl.nsel('S', 'LOC', 'Z', 0)
        >>> mapdl.d('ALL', 'UX')
        >>> mapdl.d('ALL', 'UY')
        >>> mapdl.d('ALL', 'UZ')
        >>> mapdl.nsel('S', 'LOC', 'Z', 30)
        >>> mapdl.f('ALL', 'FY', 1000)
        >>> mapdl.run('/SOLU')
        >>> mapdl.antype('STATIC')
        >>> mapdl.solve()
        >>> mapdl.finish()
        >>> mapdl.queries.uz(7)
        3.74530389e-08

        """
        return self._run_query(f"UZ({n})", integer=False)



"""Functions to download sample datasets from the pyansys data repository.
"""
import os
import shutil
import urllib.request
import zipfile

from ansys.mapdl import core as pymapdl


def get_ext(filename):
    """Extract the extension of the filename"""
    ext = os.path.splitext(filename)[1].lower()
    return ext


def delete_downloads():
    """Delete all downloaded examples to free space or update the files"""
    shutil.rmtree(pymapdl.EXAMPLES_PATH)
    os.makedirs(pymapdl.EXAMPLES_PATH)
    return True


def _decompress(filename):
    zip_ref = zipfile.ZipFile(filename, "r")
    zip_ref.extractall(pymapdl.EXAMPLES_PATH)
    return zip_ref.close()


def _get_file_url(filename, directory=None):
    if directory:
        return (
            f"https://github.com/pyansys/example-data/raw/master/{directory}/{filename}"
        )
    return f"https://github.com/pyansys/example-data/raw/master/{filename}"


def _retrieve_file(url, filename):
    # First check if file has already been downloaded
    local_path = os.path.join(pymapdl.EXAMPLES_PATH, os.path.basename(filename))
    local_path_no_zip = local_path.replace(".zip", "")
    if os.path.isfile(local_path_no_zip) or os.path.isdir(local_path_no_zip):
        return local_path_no_zip, None

    # grab the correct url retriever
    urlretrieve = urllib.request.urlretrieve

    # Perform download
    saved_file, resp = urlretrieve(url)
    shutil.move(saved_file, local_path)
    if get_ext(local_path) in [".zip"]:
        _decompress(local_path)
        local_path = local_path[:-4]
    return local_path, resp


def _download_file(filename, directory=None):
    url = _get_file_url(filename, directory)
    return _retrieve_file(url, filename)


def download_bracket():
    """Download an IGS bracket.

    Examples
    --------
    >>> from ansys.mapdl.core import examples
    >>> filename = examples.download_bracket()
    >>> filename
    '/home/user/.local/share/ansys_mapdl_core/examples/bracket.iges'

    """
    return _download_file("bracket.iges", "geometry")[0]


def download_tech_demo_data(example, filename):
    """Download Tech Demos external data."""
    example = "tech_demos/" + example
    return _download_file(filename=filename, directory=example)[0]


def download_vtk_rotor():
    """Download rotor vtk file."""
    return _download_file("rotor.vtk", "geometry")[0]


def _download_rotor_tech_demo_plot():
    """Download the rotor surface VTK file."""
    return _download_file("rotor2.vtk", "geometry")[0]


def download_example_data(filename, directory=None):
    return _download_file(filename, directory=directory)[0]



class MasterDOF:
    def m(
        self,
        node="",
        lab1="",
        nend="",
        ninc="",
        lab2="",
        lab3="",
        lab4="",
        lab5="",
        lab6="",
        **kwargs,
    ):
        """Defines master degrees of freedom for superelement generation analyses.

        APDL Command: M

        Parameters
        ----------
        node
            Node number at which master degree of freedom is defined.  If ALL,
            define master degrees of freedom at all selected nodes (NSEL).  If
            NODE = P, graphical picking is enabled and all remaining command
            fields are ignored (valid only in the GUI).  A component name may
            also be substituted for NODE.

        lab1
            Valid degree of freedom label. If ALL, use all appropriate labels.
            Structural labels:  UX, UY, or UZ (displacements); ROTX, ROTY, or
            ROTZ (rotations). Thermal labels: TEMP, TBOT, TE2, TE3, . . ., TTOP
            (temperature).  Electric labels:  VOLT (voltage).

        nend, ninc
            Define all nodes from NODE to NEND (defaults to NODE) in steps of
            NINC (defaults to 1) as master degrees of freedom in the specified
            direction.

        lab2, lab3, lab4, . . . , lab6
            Additional master degree of freedom labels.  The nodes defined are
            associated with each label specified.

        Notes
        -----
        Defines master degrees of freedom (MDOF) for superelement generation.
        If defined for other analyses, MDOF are ignored.  If used in SOLUTION,
        this command is valid only within the first load step.

        Repeat M command for additional master degrees of freedom.  The limit
        for the number of master nodes used is determined by the maximum system
        memory available.

        The substructure (ANTYPE,SUBSTR) analysis utilizes the matrix
        condensation technique to reduce the structure matrices to those
        characterized by a set of master degrees of freedom.

        Master degrees of freedom are identified by a list of nodes and their
        nodal directions.  The actual degree of freedom directions available
        for a particular node depends upon the degrees of freedom associated
        with element types (ET) at that node.  There must be some mass (or
        stress stiffening in the case of the buckling analysis) associated with
        each master degree of freedom (except for the VOLT label).  The mass
        may be due either to the distributed mass of the element or due to
        discrete lumped masses at the node.  If a master degree of freedom is
        specified at a constrained point, it is ignored.  If a master degree of
        freedom is specified at a coupled node, it should be specified at the
        prime node of the coupled set.

        Substructure analysis connection points must be defined as master
        degrees of freedom.

        This command is also valid in PREP7.
        """
        command = f"M,{node},{lab1},{nend},{ninc},{lab2},{lab3},{lab4},{lab5},{lab6}"
        return self.run(command, **kwargs)

    def mdele(
        self,
        node="",
        lab1="",
        nend="",
        ninc="",
        lab2="",
        lab3="",
        lab4="",
        lab5="",
        lab6="",
        **kwargs,
    ):
        """Deletes master degrees of freedom.

        APDL Command: MDELE

        Parameters
        ----------
        node, lab1, nend, ninc
            Delete master degrees of freedom in the Lab1 direction [M] from
            NODE to NEND (defaults to NODE) in steps of NINC (defaults to 1).
            If NODE = ALL, NEND and NINC are ignored and masters for all
            selected nodes [NSEL] are deleted.  If Lab1 = ALL, all label
            directions will be deleted.  If NODE = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may also be substituted for NODE.

        lab2, lab3, lab4, . . . , lab6
            Delete masters in these additional directions.

        Notes
        -----
        Deletes master degrees of freedom.  If used in SOLUTION, this command
        is valid only within the first load step.

        This command is also valid in PREP7.
        """
        command = (
            f"MDELE,{node},{lab1},{nend},{ninc},{lab2},{lab3},{lab4},{lab5},{lab6}"
        )
        return self.run(command, **kwargs)

    def mgen(self, itime="", inc="", node1="", node2="", ninc="", **kwargs):
        """Generates additional MDOF from a previously defined set.

        APDL Command: MGEN

        Parameters
        ----------
        itime, inc
            Do this generation operation a total of ITIMEs, incrementing all
            nodes in the set by INC each time after the first.  ITIME must be >
            1 for generation to occur.  All previously defined master degree of
            freedom directions are included in the set.  A component name may
            also be substituted for ITIME.

        node1, node2, ninc
            Generate master degrees of freedom from set beginning with NODE1 to
            NODE2 (defaults to NODE1) in steps of NINC (defaults to 1).  If
            NODE1 = ALL, NODE2 and NINC are ignored and set is all selected
            nodes [NSEL].  If NODE1 = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).

        Notes
        -----
        Generates additional master degrees of freedom from a previously
        defined set.  If used in SOLUTION, this command is valid only within
        the first load step.

        This command is also valid in PREP7.
        """
        command = f"MGEN,{itime},{inc},{node1},{node2},{ninc}"
        return self.run(command, **kwargs)

    def mlist(self, node1="", node2="", ninc="", **kwargs):
        """Lists the MDOF of freedom.

        APDL Command: MLIST

        Parameters
        ----------
        node1, node2, ninc
            List master degrees of freedom from NODE1 to NODE2 (defaults
            toNODE1) in steps of NINC (defaults to 1).  If NODE1 = ALL
            (default), NODE2 and NINC are ignored and masters for all selected
            nodes [NSEL] are listed.  If NODE1 = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may also be substituted for NODE1
            (NODE2 and NINC are ignored).

        Notes
        -----
        Lists the master degrees of freedom.
        """
        command = f"MLIST,{node1},{node2},{ninc}"
        return self.run(command, **kwargs)



class MiscellaneousLoads:
    def anpres(self, nfram="", delay="", ncycl="", refframe="", **kwargs):
        """Produces an animated sequence of the time-harmonic pressure variation

        APDL Command: ANPRES
        of an engine-order excitation in a cyclic harmonic analysis.

        Parameters
        ----------
        nfram
            Number of frame captures per cycle. Defaults to 3 times the number
            of sectors.

        delay
            Time delay (seconds) during animation. Defaults to 0.1 seconds.

        ncycl
            Number of animation cycles. Defaults to 5.

        refframe
            Reference frame for the model rotation.

            0 - Rotating reference frame (default). The model remains fixed in space and the
                pressure revolve around the model.

            1 - Stationary reference frame. The model rotates and the pressure locations remain
                fixed in space.

        Notes
        -----
        ANPRES invokes a macro which produces an animated sequence of the time-
        harmonic applied pressure in the case of a mode-superposition harmonic
        analysis (ANTYPE,HARMIC with CYCOPT,MSUP,ON). The engine-order
        excitation must also have been specified (CYCFREQ,EO). While pressure
        loads are not accepted as valid loading in a mode-superposition
        analysis (they must be applied in the modal analysis and the modal load
        vector applied in the mode-superposition analysis) you can apply them
        for the purposes of this animation.

        For RefFrame = 1 (stationary reference frame), the rotational velocity
        from the Linear Perturbation step, or the current OMEGA or CGOMGA
        value, is used to determine the rotation direction about the cyclic
        cylindrical axis, otherwise a positive rotation is assumed.

        You may use /HBC,,ON to hide overlapping pressure faces, and use
        /GLINE,,-1 to suppress the element outlines if desired.
        """
        command = f"ANPRES,{nfram},{delay},{ncycl},{refframe}"
        return self.run(command, **kwargs)

    def aport(
        self,
        portnum="",
        label="",
        kcn="",
        pres="",
        phase="",
        val1="",
        val2="",
        val3="",
        val4="",
        **kwargs,
    ):
        """Specifies input data for plane wave and acoustic duct ports.

        APDL Command: APORT

        Parameters
        ----------
        portnum
            Port number. This number is associated with an exterior port
            or interior port previously specified by the SF and BF family
            of commands, respectively. The number must be between 1 and
            50.

        Label

            * ``"PLAN"`` : Incident plane wave.
            * ``"RECT"`` : Rectangular duct.
            * ``"CIRC"`` : Circular duct.
            * ``"COAX"`` : Coaxial duct.
            * ``"LIST"`` : List the port settings. If PortNum = ALL, list the port settings for all defined ports.
            * ``"DELE"`` : Delete defined ports. If PortNum = ALL, delete all defined ports.

        kcn
            A previously-defined local (KCN >10) or global (KCN = 0)
            Cartesian coordinate system number used to specify the
            geometric properties of the duct. Defaults to the global
            Cartesian coordinate system (0). The local Z-direction must be
            the direction of wave propagation. The origin of the local
            coordinate system must be centered about the face of the duct
            port without considering symmetry.

        pres
            Zero-to-peak amplitude of the pressure. If blank, the port
            will appear as a matching impedance.

        phase
            Phase angle of the applied pressure in degrees. Defaults to 0.

        VAL1, VAL2, VAL3, VAL4
            Additional input. The meaning of VAL1 through VAL4 varies
            depending on the specified Label.  If ``label="PLAN"``:

            * ``"VAL1"`` : angle from positive X-axis to positive Y-axis
              in the local Cartesian coordinates (KCN).

            * ``"VAL2"`` : angle away from positive Z-axis in the local
              Cartesian coordinates (KCN).

            if ``label="RECT"``:

            * ``"VAL1"`` : Width of the rectangular duct.
            * ``"VAL2"`` : Height of the rectangular duct.
            * ``"VAL3"`` : Mode index for pressure variation along the
              width (defaults to 0).
            * ``"VAL4"`` : Mode index for pressure variation along the
              height (defaults to 0).

            if ``label="CIRC"``:

            * ``"VAL1"`` : Radius of the circular duct.
            * ``"VAL2"`` : Not used.
            * ``"VAL3"`` : Mode index for pressure variation along the
              azimuth (defaults to 0).
            * ``"VAL4"`` : Mode index for pressure variation along the
              radii (defaults to 0).

            if ``label="COAX"``:

            * ``"VAL1"`` : Inner radius of the coaxial duct.
            * ``"VAL2"`` : Outer radius of the coaxial duct.
            * ``"VAL3"`` : Mode index for pressure variation along the
              azimuth (defaults to 0).
            * ``"VAL4"`` : Mode index for pressure variation along the
              radii (defaults to 0).

        Notes
        -----
        Use the APORT command to launch a specified analytic acoustic mode
        into a guided duct.

        The low-order FLUID30 element does not support the higher modes in
        the coaxial duct ``label="COAX"``.

        For more information, see Specified Mode Excitation in an Acoustic
        Duct in the Acoustic Analysis Guide, and Analytic Port Modes in a
        Duct in the Mechanical APDL Theory Reference.
        """
        command = (
            f"APORT,{portnum},{label},{kcn},{pres},{phase},,{val1},{val2},{val3},{val4}"
        )
        return self.run(command, **kwargs)

    def asifile(
        self, opt="", fname="", ext="", oper="", kdim="", kout="", limit="", **kwargs
    ):
        """Writes or reads one-way acoustic-structural coupling data.

        APDL Command: ASIFILE

        Parameters
        ----------
        opt
            Command behavior option:

            WRITE  - Write the structural results to the specified file.

            READ  - Read the structural results from the specified file.

        fname
            File name and directory path of a one-way acoustic-structural
            coupling data file (248 characters maximum, including the
            characters needed for the directory path). An unspecified directory
            path defaults to the working directory; in this case, you can use
            all 248 characters for the file name (defaults to jobname).

        ext
            File name extension of the one-way acoustic-structural coupling
            data file (defaults to .asi).

        oper
            Command operation:

            NOMAP  - No mapping occurs between the structural and acoustic models when reading the
                     structural results from the specified file (default).

            MAP  - Maps the results from the structural to the acoustic model. (See "Notes".)

        kdim
            Interpolation criteria. Valid only when Oper = MAP.

        kout
            Outside region results. Valid only when Oper = MAP.

        limit
            Number of nearby nodes considered for interpolation. Valid only
            when Oper = MAP.

        Notes
        -----
        The ASIFILE command writes to, or reads from, a file containing one-way
        acoustic-structural coupling data.

        Results data on the one-way coupling interface (defined by the
        SF,,FSIN) in the structural model are written to the one-way coupling
        result data file during the structural solution.

        One-way coupling results data are read into the acoustic model as the
        velocity (harmonic) or acceleration (transient) excitation during the
        sequential acoustic solution.

        If Oper = NOMAP, both structural and acoustic models must share the
        same node number on the one-way coupling interface.

        If Oper = MAP:

        The one-way coupling interface must be defined in the acoustic model
        (SF,,FSIN) such that it corresponds to the field-surface interface
        number (FSIN) in the structural model.

        The output points are correct only if they are within the boundaries
        set via the specified input points.

        Calculations for out-of-bound points require much more processing time
        than do points that are within bounds.

        For each point in the acoustic destination mesh, the command searches
        all possible triangles in the structural source mesh to find the best
        triangle containing each point, then performs a linear interpolation
        inside this triangle. For faster and more accurate results, consider
        your interpolation method and search criteria carefully (see LIMIT).

        One-way coupling excitation can be applied to multiple frequencies or
        time steps.
        """
        command = f"ASIFILE,{opt},{fname},{ext},{oper},{kdim},{kout},{limit}"
        return self.run(command, **kwargs)

    def awave(
        self,
        wavenum="",
        wavetype="",
        opt1="",
        opt2="",
        val1="",
        val2="",
        val3="",
        val4="",
        val5="",
        val6="",
        val7="",
        val8="",
        val9="",
        val10="",
        val11="",
        val12="",
        val13="",
        **kwargs,
    ):
        """Specifies input data for an acoustic incident wave.

        APDL Command: AWAVE

        Parameters
        ----------
        wavenum
            Wave number. You specify the integer number for an acoustic
            incident wave inside or outside the model.  The number must be
            between 1 and 20.

        wavetype
            Wave type:

            PLAN - Planar incident wave

            MONO - Monopole or pulsating sphere incident wave

            DIPO - Dipole incident wave

            BACK - Back enclosed loudspeaker

            BARE - Bare loudspeaker

            STATUS - Displays the status of the acoustic wave settings if Wavenum = a number between
                     1 and 20 or ALL.

            DELE - Deletes the acoustic wave settings if Wavenum = a number between 1 and 20 or
                   ALL.

        opt1
            PRES

            PRES - Pressure

            VELO - Velocity

        opt2
            EXT

            EXT - Incident wave outside the model.

            INT - Incident wave inside the model. This option is only available for pure
                  scattered pressure formulation.

        val1, val2, val3, . . . , val13
            If Wavetype = PLAN, MONO, DIPO, BACK, or BARE:

            VAL1 - Amplitude of pressure or normal velocity to the sphere surface.

            VAL2 - Phase angle of the applied pressure or velocity (in degrees). Defaults to 0
                   degrees.

        Notes
        -----
        Use the ASOL command to activate the scattered field algorithm and the
        ASCRES command for output control with the scattered field algorithm.
        Refer to Acoustics in the Mechanical APDL Theory Reference for more
        information about pure scattered field formulation.
        """
        command = f"AWAVE,{wavenum},{wavetype},{opt1},{opt2},{val1},{val2},{val3},{val4},{val5},{val6},{val7},{val8},{val9},{val10},{val11},{val12},{val13}"
        return self.run(command, **kwargs)

    def biot(self, label="", **kwargs):
        """Calculates the Biot-Savart source magnetic field intensity.

        APDL Command: BIOT

        Parameters
        ----------
        label
            Controls the Biot-Savart calculation:

            NEW - Calculate the magnetic source field intensity (Hs) from the selected set of
                  source elements to the selected set of nodes.  Overwrite any
                  existing Hs field values.

            SUM - Calculate the Hs field from the selected set of source elements to the selected
                  set of nodes.  Accumulate with any existing Hs field values.

        Notes
        -----
        Calculates the Biot-Savart source magnetic field intensity (Hs) at the
        selected nodes from the selected source elements.  The calculation is
        done at the time the BIOT command is issued.

        Source elements include primitives described by element SOURC36, and
        coupled-field elements SOLID5, LINK68, and SOLID98.  Current conduction
        elements do not have a solved-for current distribution from which to
        calculate a source field until after the first substep.  Inclusion of a
        current conduction element Hs field will require a subsequent BIOT,SUM
        command (with SOURC36 elements unselected) and a SOLVE command.

        The units of Hs are as specified by the current EMUNIT command setting.

        This command is also valid in PREP7.
        """
        command = f"BIOT,{label}"
        return self.run(command, **kwargs)

    def dfswave(
        self,
        kcn="",
        radius="",
        psdref="",
        dens="",
        sonic="",
        incang="",
        npara="",
        sampopt="",
        **kwargs,
    ):
        """Specifies the incident planar waves with random phases for a diffuse

        APDL Command: DFSWAVE
        sound field.

        Parameters
        ----------
        kcn
            Local coordinate system:

            N - Coordinate system number. Default = 0.

            DELETE - Delete defined incident diffused planar waves.

        radius
            Radius of the reference sphere on which the incident planar waves
            are distributed with equal energy. Defaults to 50 x the half-
            maximum dimension of the structural panel.

        psdref
            Reference power spectral density. Default = 1.

        dens
            Mass density of incident planar wave media. Default = 2041 kg/m3.

        sonic
            Sound speed in incident planar wave media. Default = 343.24 m/s)

        incang
            Maximum incident angle (0o <= degree <= 180o) against the positive
            z axis in the local coordinate system KCN. Default = 0o.

        npara
            Number of divisions on the reference sphere with cutting planes
            parallel to the x-y coordinate plane of the local coordinate
            system. Default = 20.

        sampopt
            Random sampling option:

            ALL - Initializes the random generator of incident planar wave phases and samples the
                  phases at each solving frequency.

            MULT - Initializes the random generator of incident planar wave phases at the first
                   frequency and samples the phases at each solving frequency.

            MONO - Initializes the random generator of incident planar wave phases and samples the
                   phases only once at first solving frequency so that the same
                   phases are used over the whole frequency range for each
                   incident planar wave.

        Notes
        -----
        Issue the DFSWAVE command to activate a diffuse sound field. (The AWAVE
        command does not activate a diffuse sound field.)

        The SURF154 surface element must be defined on the surface of the
        structural solid element for the excitation.

        The acoustic elements and the absorbing boundary condition must be
        defined in the open acoustic domain. Do not define the acoustic domain
        on the excitation side.

        The PLST command calculates the average transmission loss for multiple
        sampling phases at each frequency over the frequency range.

        The symmetry of a panel structure cannot be used to reduce the
        simulation size, as the incident plane waves have varying random phase
        angles. The z axis of the Cartesian coordinate system (KCN) must be
        consistent with the panel’s outward normal unit vector at the center of
        the panel’s sending side.
        """
        command = (
            f"DFSWAVE,{kcn},{radius},{psdref},{dens},{sonic},{incang},{npara},{sampopt}"
        )
        return self.run(command, **kwargs)

    def fluread(
        self, fname="", ext="", kdim="", kout="", limit="", listopt="", **kwargs
    ):
        """Reads one-way Fluent-to-Mechanical APDL coupling data via a .cgns file

        APDL Command: FLUREAD
        with one-side fast Fourier transformation complex pressure peak value.

        Parameters
        ----------
        --
            Reserved.

        fname
            File name and directory path of a one-way Fluent-to-Mechanical APDL
            coupling data file (248 characters maximum, including the
            characters needed for the directory path). An unspecified directory
            path defaults to the working directory; in this case, you can use
            all 248 characters for the file name. Defaults to jobname.

        ext
            File name extension of the one-way Fluent-to-Mechanical APDL
            coupling data file. Defaults to .cgns).

        kdim
            Interpolation data for mapping. A value of 0 (default) or 2 applies
            2-D interpolation (where interpolation occurs on a surface).

        kout
            Outside region results for mapping:

            0 - Use the value(s) of the nearest region point for points outside of the region.
                This behavior is the default.

            1 - Set results extrapolated outside of the region to zero.

        limit
            Number of nearby nodes considered for mapping interpolation.
            Minimum = 5. Default = 20.

        listopt
            Type of items picked:

            (blank) - No listing (default).

            SOURCE - List the node coordinates and complex pressure values on the Fluent source side
                     during the solution.

            TARGET - List the node coordinates and complex pressure values on the mapped Mechanical
                     APDL target side during the solution.

            BOTH - List the node coordinates and complex pressure values on both the Fluent source
                   side and the mapped Mechanical APDL target side during the
                   solution.

        Notes
        -----
        The FLUREAD command reads one-way Fluent-to-Mechanical APDL coupling
        data from a .cgns file. The Fluent one-side fast Fourier transformation
        (FFT) peak complex pressure values are mapped to the Mechanical APDL
        structure model during the acoustic-structural solution at each FFT
        frequency.

        The command can be used only for the model with the acoustic elements.

        To apply complex pressure to the structure model, define the SURF154
        surface element, then define the one-way coupling interface (SF,,FSIN)
        on the element.

        You can define the solving frequency range via the HARFRQ command. The
        solver selects the FFT frequencies between the beginning and ending
        frequencies. The number of substeps is determined by the number of FFT
        frequencies over the frequency range. The number of substeps defined
        via the NSUBST command is overwritten.

        For better mapping performance, consider the following:

        Calculations for out-of-bound points require much more processing time
        than do points that are within bounds.

        For each point in the structural destination mesh, the command searches
        all possible triangles in the Fluent source mesh to find the best
        triangle containing each point, then performs a linear interpolation
        inside this triangle. For faster and more accurate results, consider
        your interpolation method and search criteria carefully. (See LIMIT.)

        It is possible to apply one-way coupling excitation to multiple
        frequencies. The one-side FFT peak complex pressure values are
        necessary to do so.
        """
        command = f"FLUREAD,{fname},{ext},{kdim},{kout},{limit},{listopt}"
        return self.run(command, **kwargs)

    def ic(self, node="", lab="", value="", value2="", nend="", ninc="", **kwargs):
        """Specifies initial conditions at nodes.

        APDL Command: IC

        Parameters
        ----------
        node
            Node at which initial condition is to be specified.  If ALL, apply
            to all selected nodes (NSEL).  If NODE = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI). A component name may be substituted for NODE.

        lab
            Degree-of-freedom label for which the initial condition is to be
            specified.  If ALL, use all appropriate labels.

        value
            Initial value of the degree of freedom (first-order value).
            Defaults to the program default for that degree of freedom (0.0 for
            structural analysis, TUNIF for thermal analysis, etc.). Values are
            in the nodal coordinate system and in radians for rotational
            degrees of freedom.

        value2
            Second-order degree of freedom value, mainly used to specify
            initial structural velocity.  Defaults to the program default for
            that degree of freedom (0.0 for structural analysis).  Values are
            in the nodal coordinate system and in radians/time for rotational
            degrees of freedom.

        nend, ninc
            Specifies the same initial condition values at the range of nodes
            from NODE to NEND (defaults to NODE), in steps of NINC (defaults to
            1).

        Notes
        -----
        The IC command specifies initial conditions, which are the initial
        values of the specified degrees of freedom. It is valid only for a
        static analysis and full method transient analysis (TIMINT,ON and
        TRNOPT,FULL). For the transient, the initial value is specified at the
        beginning of the first load step, that is, at time = 0.0.

        Initial conditions should always be step applied (KBC,1) and not
        ramped.

        If constraints (D, DSYM, etc.) and initial conditions are applied at
        the same node, the constraint specification overrides. Exercise caution
        when specifying constraints. The degree-of-freedom values start from
        zero, or the first value given in the table when table name is
        specified. To match the nonzero initial condition value with the
        initial value for degree-of-freedom constraint, use a table for the
        degree-of-freedom constraint.

        For thermal analyses, any TUNIF specification should be applied before
        the IC command; otherwise, the TUNIF specification is ignored.  If the
        IC command is input before any TUNIF specification, use the ICDELE
        command and then reissue any TUNIF specification and then follow with
        the IC command.

        When issuing the IC command for elements SOLID278 Layered Thermal Solid
        and SOLID279 Layered Thermal Solid with through-the-thickness degrees
        of freedom (KEYOPT(3) = 2), layers are always interpolated linearly
        based on the location of the degrees of freedom.

        Define consistent initial conditions. For example, if you define an
        initial velocity at a single degree of freedom, the initial velocity at
        every other degree of freedom will be 0.0, potentially leading to
        conflicting initial conditions. In most cases, you should define
        initial conditions at every unconstrained degree of freedom in your
        model. If you define an initial condition for any degree of freedom at
        the pilot node of a rigid body (see Modeling Rigid Bodies in the
        Contact Technology Guide for the definition of rigid body), then the
        same initial condition must also be defined for the same degree of
        freedom on all other nodes of the rigid body.

        After a solution has been performed, the specified initial conditions
        are overwritten by the actual solution and are no longer available. You
        must respecify them if you want to perform a subsequent analysis. You
        may want to keep a database file saved prior to the first solution for
        subsequent reuse.

        If you use the CDWRITE command to archive your model, first-order
        values (initial displacements, temperatures, etc.) specified via the IC
        command are not written to the archive file; however, second-order
        (structural velocity) terms are written.

        This command is also valid in PREP7.
        """
        command = f"IC,{node},{lab},{value},{value2},{nend},{ninc}"
        return self.run(command, **kwargs)

    def icdele(self, **kwargs):
        """Deletes initial conditions at nodes.

        APDL Command: ICDELE

        Notes
        -----
        Deletes all initial conditions previously specified with the IC command
        at all nodes.

        This command is also valid in PREP7.
        """
        command = f"ICDELE,"
        return self.run(command, **kwargs)

    def iclist(self, node1="", node2="", ninc="", lab="", **kwargs):
        """Lists the initial conditions.

        APDL Command: ICLIST

        Parameters
        ----------
        node1, node2, ninc
            List initial conditions for nodes NODE1 to NODE2 (defaults to
            NODE1) in steps of NINC (defaults to 1).  If NODE1 = ALL (default),
            NODE2 and NINC are ignored and initial conditions for all selected
            nodes [NSEL] are listed.  If NODE1 = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).  A component name may be substituted for NODE1 (NODE2 and
            NINC are ignored).

        lab
            Velocity key:

            DISP - Specification is for first order degree of freedom value (displacements,
                   temperature, etc.) (default).

            VELO - Specification is for second order degree of freedom value (velocities).

        Notes
        -----
        Lists the initial conditions specified by the IC command.  Listing
        applies to all the selected nodes [NSEL] and DOF labels.  ICLIST is not
        the same as the DLIST command.  All the initial conditions including
        the default conditions are listed for the selected nodes.

        This command is valid in any processor.
        """
        command = f"ICLIST,{node1},{node2},{ninc},{lab}"
        return self.run(command, **kwargs)

    def icrotate(
        self,
        node="",
        omega="",
        x1="",
        y1="",
        z1="",
        x2="",
        y2="",
        z2="",
        vx="",
        vy="",
        vz="",
        accel="",
        **kwargs,
    ):
        """Specifies initial velocity at nodes as a sum of rotation about an axis and translation.

        APDL Command: ICROTATE

        Parameters
        ----------
        NODE
            Node at which the initial velocity is to be specified. If ALL,
            apply to all selected nodes NSEL.  A component name may be
            input for NODE.

        OMEGA
            Scalar rotational velocity about the rotational axis.

        X1, Y1, Z1
            Coordinates (in the global Cartesian coordinate system) of the
            beginning point of the rotational axis vector.

        X2, Y2, Z2
            Coordinates (in the global Cartesian coordinate system) of the
            end point of the rotational axis vector.

        Vx
            Initial translational velocity in direction x of the nodal
            coordinate system.

        Vy
            Initial translational velocity in direction y of the nodal
            coordinate system.

        Vz
            Initial translational velocity in direction z of the nodal
            coordinate system.

        accel
            Key to initialize acceleration due to centrifugal effects:

            * ``""`` : (blank) Do not initialize acceleration (default).
            * ``"CENT"`` : Initialize acceleration due to centrifugal
              effects along with the initial velocity.

        Notes
        -----
        The ICROTATE command specifies initial velocity for all
        translational degrees of freedom of the specified nodes. The
        velocity value is a combination of velocity due to rotation about
        an axis and translation.
        """
        command = f"ICROTATE,{node},{omega},{x1},{y1},{z1},{x2},{y2},{z2},{vx},{vy},{vz},{accel}"
        return self.run(command, **kwargs)

    def mrpm(self, val1="", **kwargs):
        """Defines the revolutions per minute (RPM) for a machine rotation.

        APDL Command: MRPM

        Parameters
        ----------
        val1
            The RPM value (no default).

        Notes
        -----
        A different RPM value can be defined at each load step. The RPM
        value is used to postprocess the equivalent radiated power from
        the structural surface (the PRAS and PLAS commands) or the
        radiated sound power level (the PRFAR and PLFAR commands).
        """
        return self.run(f"MRPM,{val1}", **kwargs)

    def outpr(self, item="", freq="", cname="", **kwargs):
        """Controls the solution printout.

        APDL Command: OUTPR

        Parameters
        ----------
        item
            Item for print control:

            BASIC - Basic quantities (nodal DOF solution, nodal reaction loads, and element
                    solution) (default).

            NSOL - Nodal DOF solution.

            RSOL - Nodal reaction loads.

            ESOL - Element solution.

            NLOAD - Element nodal loads. When nonlinear stabilization is active, the stabilization
                    force/moments are also printed.

            SFOR - Stabilization force/moment at the applicable nodes (valid only when nonlinear
                   stabilization is active).

            VENG - Element energies. When nonlinear stabilization is active, the energy
                   dissipation due to stabilization is also printed.

            V - Nodal velocity (applicable to structural transient analysis  only
                (ANTYPE,TRANS)).

            A - Nodal acceleration (applicable to structural transient analysis only
                (ANTYPE,TRANS)).

            ALL - All of the above solution items.

        freq
            Print solution for this item every Freqth (and the last) substep of
            each load step.  If -n, print up to n equally spaced solutions
            (only applies to static or full transient analyses when automatic
            time stepping is enabled).  If NONE, suppress all printout for this
            item for this load step.  If ALL, print solution for this item for
            every substep.  If LAST, print solution for this item only for the
            last substep of each load step.  For a modal analysis, use NONE or
            ALL.

        cname
            Name of the component, created with the CM command, defining the
            selected set of nodes or elements for which this specification is
            active.  If blank, the set is all entities.

        Notes
        -----
        Controls the solution items to be printed, the frequency with which
        they are printed (in static, transient, or full harmonic analyses), and
        the set of nodes or elements to which this specification applies (in
        static, transient, or full harmonic analyses).  An item is associated
        with either a node (NSOL, RFORCE, V, and A items) or an element (all of
        the remaining items).  The specifications are processed in the order
        that they are input.  Up to 50 specifications (OUTPR and  OUTRES
        commands combined) may be defined.  Use OUTPR,STAT to list the current
        specifications and use OUTPR,ERASE to erase all the current
        specifications.

        As described above, OUTPR writes some or all items (depending on
        analysis type) for all elements.  To restrict the solution printout,
        use OUTPR to selectively suppress (Freq = NONE) the writing of solution
        data, or first suppress the writing of all solution data
        (OUTPR,ALL,NONE) and then selectively turn on the writing of solution
        data with subsequent OUTPR commands.

        If the generalized plane strain feature is active and OUTPR is issued,
        the change of fiber length at the ending point during deformation and
        the rotation of the ending plane about X and Y during deformation will
        be printed if any displacement at the nodes is printed. The reaction
        forces at the ending point will be printed if any reaction force at the
        nodes is printed.

        Nodal reaction loads (Item = RSOL) are processed according to the
        specifications listed for the PRRSOL command.

        Result printouts for interactive sessions are suppressed for models
        with more than 10 elements.

        This command is also valid in PREP7.
        """
        command = f"OUTPR,{item},{freq},{cname}"
        return self.run(command, **kwargs)

    def outres(self, item="", freq="", cname="", nsvar="", dsubres="", **kwargs):
        """Controls the solution data written to the database.

        APDL Command: OUTRES

        Parameters
        ----------
        item
            Results item for database and file write control:

            ALL - All solution items except LOCI and SVAR. This behavior is the default.

            CINT - All available results generated by the CINT command

            ERASE - Resets OUTRES specifications to their default values.

            STAT - Lists the current OUTRES specifications.

            BASIC - Write only NSOL, RSOL, NLOAD, STRS, FGRAD, and FFLUX records to the results
                    file and database.

            NSOL - Nodal DOF solution.

            RSOL - Nodal reaction loads.

            V - Nodal velocity (applicable to structural full transient analysis only
                (ANTYPE,TRANS)).

            A - Nodal acceleration (applicable to structural full transient analysis only
                (ANTYPE,TRANS)).

            ESOL - Element solution (includes all items following):

            NLOAD - Element nodal, input constraint, and force loads (also used with the /POST1
                    commands PRRFOR, NFORCE, and FSUM to calculate reaction
                    loads).

            STRS - Element nodal stresses.

            EPEL - Element elastic strains.

            EPTH - Element thermal, initial, and swelling strains.

            EPPL - Element plastic strains.

            EPCR - Element creep strains.

            EPDI - Element diffusion strains.

            FGRAD - Element nodal gradients.

            FFLUX - Element nodal fluxes.

            LOCI - Integration point locations.

            SVAR - State variables (used only by UserMat).

            MISC - Element miscellaneous data (SMISC and NMISC items of the ETABLE command).

        freq
            Specifies how often (that is, at which substeps) to write the
            specified solution results item. The following values are valid:

        cname
            The name of the component, created with the CM command, defining
            the selected set of elements or nodes for which this specification
            is active.  If blank, the set is all entities.  A component name is
            not allowed with the ALL, BASIC, or RSOL items.

        --
            Reserved for future use.

        nsvar
            The number of user-defined state variables (TB,STATE) to be written
            to the results file. Valid only when Item = SVAR and user-defined
            state variables exist. The specified value cannot exceed the total
            number of state variables defined; if no value is specified, all
            user-defined state variables are written to the results file. This
            argument acts on all sets of user-defined state variables that
            exist for the model.

        dsubres
            Specifies whether to write additional results in Jobname.DSUB
            during a substructure or CMS use pass in transient or harmonic
            analysis.

            Blank - Write the nodal DOF solution in Jobname.DSUB (default).

            ALL - In addition to the nodal DOF solution, also write necessary data to compute
                  quantities using nodal velocity and nodal acceleration
                  (damping force, inertial force, kinetic energy, etc.) in the
                  subsequent expansion pass. For more information, see Step 3:
                  Expansion Pass in the Substructuring Analysis Guide.

        Notes
        -----
        The OUTRES command allows you to specify the following:

        The solution item (Item) to write to the database (and to the reduced
        displacement and results files)

        The frequency (Freq) at which the solution item is written (applicable
        to static, transient, or full harmonic analyses)

        The set of elements or nodes (Cname) to which your specification
        applies.
        """
        command = f"OUTRES,{item},{freq},{cname},{nsvar},{dsubres}"
        return self.run(command, **kwargs)

    def rescontrol(
        self,
        action="",
        ldstep="",
        frequency="",
        maxfiles="",
        maxtotalfiles="",
        filetype="",
        **kwargs,
    ):
        """Controls file writing for multiframe restarts.

        APDL Command: RESCONTROL

        Parameters
        ----------
        action
            Command action. Valid options are:

            DEFINE
                Issuing the command specifies how frequently the ``.Xnnn`` restart files are
                written for a load step (default).

            FILE_SUMMARY
                Issuing the command prints the substep and load step information for all ``.Xnnn``
                files for the current jobname in the current
                directory. If this option is specified, all other
                arguments are ignored.

            STATUS
                Issuing the command lists the current status in the tables of restart controls
                specified previously by RESCONTROL. If this option is
                specified, all other arguments are ignored.

            NORESTART
                Issuing the command cleans up some of the restart files after a Distributed
                ANSYS solution. The host process will not have the
                following files in the working directory at the end of
                the run: ``.ESAV``, ``.OSAV``, ``.Xnnn``, ```, `.LDHI``. The slave
                processes will not have the following files in the
                working directory at the end of the run: ``.ESAV``, ``.OSAV``,
                ``.Xnnn``, ``.RST`` (or ``.RTH``, etc.). Some of the restart files
                are never written, some are removed upon leaving ``/SOLU``
                (for example, upon FINISH), and some are removed upon
                exiting the program.

                This option is useful for cleaning up files written by all of the Distributed
                ANSYS processes, particularly when you know that these restart files will not
                be needed later on. If this option is specified, all other arguments are ignored.

                If this option is used in shared-memory parallel ANSYS, most of the restart
                files in the working directory are removed. It
                has the same effect as issuing ``RESCONTROL,,NONE``.

            LINEAR
                Issuing the command specifies the same actions as Action = DEFINE. However,
                this option is intended for linear static applications.
                For a linear static analysis, the restart capability is
                normally not needed. However, it is typically needed when
                a subsequent linear perturbation analysis is desired. By
                default, none of the restart files are written for a
                linear static analysis.

            DELETE
                Delete the restart control specification corresponding to the ``Ldstep`` label on a
                previous ``RESCONTROL,DEFINE`` command.

        ldstep
            Specifies how the ``.Xnnn`` files are written for the specified load
            steps. This option also affects how often the load history
            information is written to the ``.LDHI`` file.

            ALL
                Write the ``.Xnnn`` files at the same substep Frequency for all load steps; write
                the load history information to the ``.LDHI`` file for all load
                steps.

            LAST
                Write the ``.Xnnn`` files for the last load step only; write load history
                information to the ``.LDHI`` file for the last load step only.
                This option is the default for nonlinear static and full
                transient analyses. The remaining arguments are ignored.

            N
                Number that indicates how often the ``.Xnnn`` file is written.

                Input a positive number to write the ``.Xnnn`` files at the substep ``Frequency``
                indicated only for load step N.
                Other load steps will be written at the default substep frequency or at a frequency
                defined by a previous ``RESCONTROL`` specification.
                Load history information is written to the ``.LDHI`` file only for load steps N.

                Specifying a negative number (-N) to write the ``.Xnnn`` files for every Nth load step
                at the specified substep Frequency. The load
                history information is written to the ``.LDHI`` file
                every Nth load step. This option is suitable for
                restart applications in which more than a few
                hundred load steps are required. Compared to the
                ALL and positive N options, it can save disk
                space since the ``.LDHI`` file is smaller and fewer
                ``.Xnnn`` files are written.

                If Ldstep = -N, all other Ldstep options specified by ``RESCONTROL`` are
                ignored and the program follows the -N option (write load history information every Nth load step).
                If you want to change this pattern, issue ``RESCONTROL,DELETE, -N`` and then issue another
                ``RESCONTROL`` command with the desired Ldstep option.

            NONE
                No multiframe restart files (``.RDB`` [restart database file], ``.LDHI`` [load history file],
                ``.Xnnn``) are created. This option is the default for mode-superposition analyses.
                The remaining arguments are ignored.

                For nonlinear static, linear static, and full transient analyses, this option
                allows a restart to be done at the last or abort
                point using the same procedure as in ANSYS 5.5 or
                earlier (using the ``.EMAT``, ``.ESAV`` or ``.OSAV``, and ``.DB``
                files).

        frequency
            Frequency at which the ``.Xnnn`` files are written at the substep
            level.

            NONE
                Do not write any ``.Xnnn`` files for this load step.

            LAST
                Write the ``.Xnnn`` files for the last substep of the load step only (default for
                nonlinear static and full transient analyses).

            N
                If ``N`` is positive, write the ``.Xnnn`` file every Nth substep of a load step. If ``N``
                is negative, write ``N`` equally spaced ``.Xnnn`` files within a load step.

                In nonlinear static and full transient analyses, negative ``N`` is valid only when ``AUTOTS,ON``.
                In mode-superposition analyses, negative ``N`` is always valid.

        maxfiles
            Maximum number of ``.Xnnn`` files to save for Ldstep.

            \-1
                Overwrite existing ``.Xnnn`` files (default). The total maximum number of ``.Xnnn``
                files for one run is 999. If this number is reached before the analysis is complete,
                the program will reset the ``.Xnnn`` file numbering back to 1 and continue to write
                ``.Xnnn`` files; the program keeps the newest 999 restart files and overwrites the
                oldest restart files.

            0
                Do not overwrite any existing ``.Xnnn`` files. The total maximum number of ``.Xnnn``
                files for one run is 999. If this number is reached before the
                analysis is complete, the analysis continues but no longer
                writes any ``.Xnnn`` files.

            N
                Maximum number of ``.xnnn`` files to keep for each load step. When ``N.xnnn`` files have
                been written for a load step, the program overwrites the first ``.xnnn`` file of that load step
                for subsequent substeps.

                .. warning:: ``N`` must be <= 999. If a total of 999 restart files is reached before
                   the analysis is complete, the analysis continues but writes no additional ``.xnnn`` files.

        maxtotalfiles
            Total number of restart files to keep. ``Default = 999`` for ``.xnnn`` files and 99 for ``.rdnn`` files. This
            option is valid only when ``MAXFILES`` = -1 (default).
            Valid ``maxtotalfiles`` values are 1 through 999 for ``.xnnn`` files, and 1 through 99 for ``.rdnn`` files.

            When the total number of restart files written exceeds ``maxtotalfiles``, the program resets the
            ``.xnnn`` or ``.rdnn`` file numbering back to 1 and continues to write ``.xnnn`` or ``.rdnn`` files. The
            newest files are retained and the oldest files are overwritten.

            The ``maxtotalfiles`` value specified applies to all subsequent load steps. To reset it to the default,
            reissue the command with ``maxtotalfiles`` = 0 or some negative value.

            If ``maxtotalfiles`` is set to different values at different load steps, and if the value of ``maxtotalfiles``
            specified in the prior load step is larger than that of the current load step, the program can
            only overwrite the current number of maximum restart files up to the number ``maxtotalfiles``
            currently specified (which is smaller than the previous number).

            The recommended way to control the maximum number of restart files is to specify ``maxtotalfiles``
            at the first load step and not vary it in subsequent load steps. Also, ``maxtotalfiles`` is best used
            when Ldstep = -N or ALL.

        filetype
            The type of restart file to be controlled by this command. Valid only when Action = DEFINE:

            * **XNNN**: Control ``.xnnn`` files (default).
            * **RDNN**: Control ``.rdnn`` remeshing database files. Needed only for a nonlinear mesh adaptivity analysis.

        Notes
        -----

        **COMMAND DEFAULT**

        If the ``RESCONTROL`` command is not issued during a structural analysis, the ``.RDB`` and ``.LDHI`` files are
        written as described in Restarting an Analysis.

        **For nonlinear static and full transient analysis:**

        The default behavior is multiframe restart: ``RESCONTROL,DEFINE,LAST,LAST``
        The ``.xnnn`` file is written at the last substep of the last load step. An ``.rnnn`` file is also written at
        the iteration prior to the abort point of the run if a ``jobname.abt`` file was used (or the **Stop** button
        was pressed in the GUI), or if the job terminated because of a failure to reach convergence or some
        other solution error. No information at the aborted substep is saved to the ``.xnnn`` file.

        **For nonlinear mesh adaptivity analysis:**

        The default behavior for ``.rdnn`` files written is: ``RESCONTROL,DEFINE,ALL,LAST,,,,RDNN``
        The ``.rdnn`` file is written at the last remesh of every load step by default.
        The ``.rdnn`` and ``.rnnn`` files interact with each other. Generally, ``.rdnn`` file writing is superior to
        that of ``.rnnn`` file writing. For example, if no RESCONTROL,DEFINE command is issued, the default
        behavior is that both ``.rdnn`` and ``.rnnn`` files are written at the last occurrence of every load step
        (equivalent to ``RESCONTROL,DEFINE,ALL,LAST`` and ``RESCONTROL,DEFINE,ALL,LAST,,,,RDNN``)


        **NOTES**

        ``RESCONTROL`` sets up the restart parameters for a multiframe restart, enabling you to restart an analysis
        from any load step and substep for which there is an ``.xnnn`` file. You can perform a multiframe restart
        for static and transient (full or mode-superposition method) analyses only. For more information about
        multiframe restarts and descriptions of the contents of the files used, see *Restarting an Analysis* in the
        *Basic Analysis Guide*.

        **Syntax**

        Multiframe restart files are generically indicated here as ``.xnnn`` files. They correspond to .rnnn
        files for nonlinear static and full transient analyses, and ``.mnnn`` files for mode-superposition
        analyses.
        Remeshing database files are indicated as ``.rdnn`` files. This type of restart file is needed only
        after remeshing during a *nonlinear mesh adaptivity* analysis.
        When ``Action = DEFINE``, the specified Filetype determines the type of file (``.xnnn`` or
        ``.rdnn``) controlled by this command.

        **Number of Restart Files Allowed**

        The total number of restart files for any analysis cannot exceed 999 (for example, ``jobname.r001``
        to ``jobname.r999``).
        The total number of remeshing database files cannot exceed 99 (for example, ``jobname.rd01``
        to ``jobname.rd99``).

        **Considerations for Nonlinear Mesh Adaptivity Analysis**

        To control both ``.xnnn`` and ``.rdnn`` file writing (``Filetype = XNNN`` and ``Filetype = RDNN``,
        respectively), separate ``RESCONTROL`` commands are necessary.
        ``Action = NORESTART`` and ``Ldstep = NONE`` are not valid and will cause the analysis to fail.
        ``Ldstep = -N`` is not valid for controlling ``.xnnn`` files.

        **Limiting the Number of Files Saved**

        If you have many substeps for each load step and are writing ``.xnnn`` files frequently, you may
        want to set ``maxfiles`` to limit the number of ``.xnnn`` files saved, as they can fill your disk
        quickly.
        You can specify ``maxfiles`` and ``Frequency`` for individual load steps. These arguments take
        on the default value or the value defined by ``RESCONTROL,,ALL,Frequency,maxfiles`` if they
        are not explicitly defined for a specific load step.
        When ``.xnnn`` files are written over many load steps, you may want to further limit the number
        of ``.xnnn`` files by setting ``maxtotalfiles``.

        **Maximum Number of Load Steps**

        You can specify a maximum of ten load steps; that is, you can issue the ``RESCONTROL,,N``
        command a maximum of ten times. Specified load steps cannot be changed in a restart.

        **Specifying Ldstep = LAST or Ldstep = -N**

        The program accepts only one occurrence of ``RESCONTROL`` with ``Ldstep = LAST``. If you issue
        ``RESCONTROL,,LAST,Frequency,maxfiles`` multiple times, the last specification overwrites
        the previous one.
        The program accepts only one occurrence of ``RESCONTROL`` with a negative ``Ldstep`` value
        (``RESCONTROL,,N`` where ``N`` is a negative number). If you issue ``RESCONTROL`` multiple times
        with a negative ``Ldstep`` value, the last specification overwrites the previous one.

        """
        command = f"RESCONTROL, {action}, {ldstep}, {frequency}, {maxfiles}, , {maxtotalfiles}, {filetype}"
        return self.run(command, **kwargs)

    def sbclist(self, **kwargs):
        """Lists solid model boundary conditions.

        APDL Command: SBCLIST

        Notes
        -----
        Lists all solid model boundary conditions for the selected solid model
        entities.  See also DKLIST, DLLIST, DALIST, FKLIST, SFLLIST, SFALIST,
        BFLLIST, BFALIST, BFVLIST, and BFKLIST to list items separately.

        This command is valid in any processor.
        """
        command = f"SBCLIST,"
        return self.run(command, **kwargs)

    def sbctran(self, **kwargs):
        """Transfers solid model loads and boundary conditions to the FE model.

        APDL Command: SBCTRAN

        Notes
        -----
        Causes a manual transfer of solid model loads and boundary conditions
        to the finite element model.  Loads and boundary conditions on
        unselected keypoints, lines, areas, and volumes are not transferred.
        Boundary conditions and loads will not be transferred to unselected
        nodes or elements.  The SBCTRAN operation is also automatically done
        upon initiation of the solution calculations [SOLVE].

        This command is also valid in PREP7.
        """
        command = f"SBCTRAN,"
        return self.run(command, **kwargs)

    def wsprings(self, **kwargs):
        """Creates weak springs on corner nodes of a bounding box of the currently

        APDL Command: WSPRINGS
        selected elements.

        Notes
        -----
        WSPRINGS invokes a predefined ANSYS macro that is used during the
        import of loads from the ADAMS program into the ANSYS program. WSPRINGS
        creates weak springs on the corner nodes of the bounding box of the
        currently selected elements. The six nodes of the bounding box are
        attached to ground using COMBIN14 elements. The stiffness is chosen as
        a small number and can be changed by changing the real constants of the
        COMBIN14 elements. This command works only for models that have a
        geometric extension in two or three dimensions. One dimensional
        problems (pure beam in one axis) are not supported.

        For more information on how WSPRINGS is used during the transfer of
        loads from the ADAMS program to ANSYS, see Import Loads into ANSYS in
        the Substructuring Analysis Guide.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"WSPRINGS,"
        return self.run(command, **kwargs)



class RadiationMat:
    def emis(self, mat="", evalu="", **kwargs):
        """Specifies emissivity as a material property for the Radiation Matrix

        APDL Command: EMIS
        method.

        Parameters
        ----------
        mat
            Material number associated with this emissivity (500 maximum).
            Defaults to 1.

        evalu
            Emissivity for this material (0.0 < EVALU   1.0).  Enter a very
            small number for zero.

        Notes
        -----
        Specifies emissivity as a material property for the Radiation Matrix
        method. This material property can then be associated with each
        element.
        """
        command = f"EMIS,{mat},{evalu}"
        return self.run(command, **kwargs)

    def geom(self, k2d="", ndiv="", **kwargs):
        """Defines the geometry specifications for the radiation matrix

        APDL Command: GEOM
        calculation.

        Parameters
        ----------
        k2d
            Dimensionality key:

            0 - 3-D geometry (default)

            1 - 2-D geometry (plane or axisymmetric)

        ndiv
            Number of divisions in an axisymmetric model.  Used only with K2D =
            1.  Defaults to 0 (2-D plane).  The 2-D model is internally
            expanded to a 3-D model based on the number of divisions specified
            (6   NDIV   90).  For example, NDIV of 6 is internally represented
            by six 60° sections.
        """
        command = f"GEOM,{k2d},{ndiv}"
        return self.run(command, **kwargs)

    def mprint(self, key="", **kwargs):
        """Specifies that radiation matrices are to be printed.

        APDL Command: MPRINT

        Parameters
        ----------
        key
            Print key:

            0 - Do not print matrices.

            1 - Print matrices.

        Notes
        -----
        Specifies that the element and node radiation matrices are to be
        printed when the WRITE command is issued.  If KEY = 1, form factor
        information for each element will also be printed.
        """
        command = f"MPRINT,{key}"
        return self.run(command, **kwargs)

    def space(self, node="", **kwargs):
        """Defines a space node for radiation using the Radiation Matrix method.

        APDL Command: SPACE

        Parameters
        ----------
        node
            Node defined to be the space node.

        Notes
        -----
        A space node is required in an open system to account for radiation
        losses.

        If using SPACE with the ANSYS Multi-field solver (MFS),  you must
        capture this command in the command file using MFCMMAND. This step is
        necessary because at the end of each field computation, this command is
        unset.
        """
        command = f"SPACE,{node}"
        return self.run(command, **kwargs)

    def vtype(self, nohid="", nzone="", **kwargs):
        """Specifies the viewing procedure used to determine the form factors for

        APDL Command: VTYPE
        the Radiation Matrix method.

        Parameters
        ----------
        nohid
            Type of viewing procedure:

            0 - Hidden procedure.

            1 - Non-hidden (faster, but less general) procedure.

        nzone
            Number of sampling zones for the hidden procedure (100 maximum for
            3-D, 1000 maximum for 2-D).  Defaults to 20 for 3-D, 200 for 2-D.
            Number of points is ``2*NZONE`` for 2-D and ``2*NZONE*(NZONE+1)`` for 3-D.
        """
        command = f"VTYPE,{nohid},{nzone}"
        return self.run(command, **kwargs)

    def write(self, fname="", **kwargs):
        """Writes the radiation matrix file.

        APDL Command: WRITE

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        Notes
        -----
        Writes radiation matrix file (File.SUB) for input to the substructure
        thermal "use" pass.  Subsequent WRITE operations to the same file
        overwrite the file.
        """
        command = f"WRITE,{fname}"
        return self.run(command, **kwargs)



# Importing logging
import logging
import os

import appdirs

from ansys.mapdl.core.logging import Logger

LOG = Logger(level=logging.ERROR, to_file=False, to_stdout=True)
LOG.debug("Loaded logging module as LOG")

_LOCAL_PORTS = []

# Per contract with Sphinx-Gallery, this method must be available at top level
try:
    from pyvista.utilities.sphinx_gallery import _get_sg_image_scraper

    _HAS_PYVISTA = True
except ModuleNotFoundError:  # pragma: no cover
    LOG.debug("The module 'Pyvista' is not installed.")
    _HAS_PYVISTA = False

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata

__version__ = importlib_metadata.version(__name__.replace(".", "-"))

from ansys.mapdl.core import examples
from ansys.mapdl.core.convert import convert_apdl_block, convert_script
from ansys.mapdl.core.launcher import (
    change_default_ansys_path,
    close_all_local_instances,
    find_ansys,
    get_ansys_path,
    launch_mapdl,
)
from ansys.mapdl.core.mapdl_grpc import MapdlGrpc as Mapdl
from ansys.mapdl.core.misc import Information, Report, _check_has_ansys
from ansys.mapdl.core.pool import LocalMapdlPool
from ansys.mapdl.core.theme import MapdlTheme

_HAS_ANSYS = _check_has_ansys()

# Setup data directory
try:
    USER_DATA_PATH = appdirs.user_data_dir("ansys_mapdl_core")
    if not os.path.exists(USER_DATA_PATH):  # pragma: no cover
        os.makedirs(USER_DATA_PATH)

    EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")
    if not os.path.exists(EXAMPLES_PATH):  # pragma: no cover
        os.makedirs(EXAMPLES_PATH)

except:  # pragma: no cover
    pass


# override default launcher when on pyansys.com
if "ANSJUPHUB_VER" in os.environ:
    from ansys.mapdl.core.jupyter import launch_mapdl_on_cluster as launch_mapdl


BUILDING_GALLERY = False



class Special:
    def cvar(self, ir="", ia="", ib="", itype="", datum="", name="", **kwargs):
        """Computes covariance between two quantities.

        APDL Command: CVAR

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previous
            variable, the previous variable will be overwritten with this
            result.

        ia, ib
            Reference numbers of the two variables to be operated on.  If only
            one, leave IB blank.

        itype
            Defines the type of response PSD to be calculated:

            0,1 - Displacement (default).

            2 - Velocity.

            3 - Acceleration.

        datum
            Defines the reference with respect to which covariance is to be
            calculated:

            1 - Absolute value.

            2 - Relative to base (default).

        name
            Thirty-two character name for identifying the variable on listings
            and displays.  Embedded blanks are compressed upon output.

        Notes
        -----
        This command computes the covariance value for the variables referenced
        by the reference numbers IA and IB.  If DATUM = 2, the variable
        referenced by IR will contain the individual modal contributions (i.e.,
        the dynamic or relative values).  If DATUM = 1, the variable referenced
        by IR will contain the modal contributions followed by the
        contributions of pseudo-static and covariance between dynamic and
        pseudo-static responses. File.PSD must be available for the
        calculations to occur.
        """
        command = f"CVAR,{ir},{ia},{ib},{itype},{datum},{name}"
        return self.run(command, **kwargs)

    def pmgtran(
        self,
        fname="",
        freq="",
        fcnam1="",
        fcnam2="",
        pcnam1="",
        pcnam2="",
        ecnam1="",
        ccnam1="",
        **kwargs,
    ):
        """Summarizes electromagnetic results from a transient analysis.

        APDL Command: PMGTRAN

        Parameters
        ----------
        fname
            File name (8 characters maximum) to which tabular data and plot
            files will be written.  Must be enclosed in single quotes when the
            command is manually typed in.  Defaults to MG_TRNS.  The data file
            extension is .OUT and the plot file extension is .PLT.

        freq
            Frequency of solution output.  Defaults to 1.  Every FREQth
            solution on the results file is output.

        fcnam1, fcnam2
            Names of element components for force calculation.  Must be
            enclosed in single quotes when the command is manually typed in.

        pcnam1, pcnam2
            Names of element components for power loss calculation.  Must be
            enclosed in single quotes when the command is manually typed in.

        ecnam1, ccnam1
            Names of element components for energy and total current
            calculations, respectively.  Must be enclosed in single quotes when
            the command is manually typed in.

        Notes
        -----
        PMGTRAN invokes an ANSYS macro which calculates and summarizes
        electromagnetic results from a transient analysis.  The results are
        summarized by element components and listed on the screen as well as
        written to a file (Fname.OUT).  Also, graph plots of results as a
        function of time are created and written to a file (Fname.PLT) for use
        in the DISPLAY program.

        Two components may be selected for the summary of electromagnetic
        forces (see FMAGSUM), two for power loss, and one each for stored
        energy (see SENERGY) and total current (see CURR2D).  See the
        referenced commands for other restrictions.

        PMGTRAN is restricted to MKSA units.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"PMGTRAN,{fname},{freq},{fcnam1},{fcnam2},{pcnam1},{pcnam2},{ecnam1},{ccnam1}"
        return self.run(command, **kwargs)

    def rcyc(self, ir="", ia="", sector="", name="", **kwargs):
        """Calculates cyclic results for a mode-superposition harmonic solution.

        APDL Command: RCYC

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]). If this number is the same as for a previous
            variable, the previous variable will be overwritten with this
            result.

        ia
            Reference number of the variable to be operated on.

        sector
            Sector number to calculate the results for.

        name
            Thirty-two character name identifying the variable on listings and
            displays. Embedded blanks are compressed for output.

        Notes
        -----
        This command calculates the harmonic response in the sector specified
        by SECTOR for the variable referenced by the reference number IA. Only
        component values for IA are valid (no principles or sums). The variable
        specified by IR will contain the harmonic solution. Jobname.RFRQ from
        the cyclic mode-superposition harmonic solve and Jobname.RST or
        Jobname.RSTP from the cyclic modal solve must be available for the
        calculations to occur. The Jobname must be the same for the cyclic
        modal solve and the cyclic mode-superposition harmonic solve.

        For SECTOR > 1, the result is in the nodal coordinate system of the
        base sector, and it is rotated to the expanded sector’s location. Refer
        to Using the /CYCEXPAND Command in the Cyclic Symmetry Analysis Guide
        for more information.

        See also Mode-Superposition Harmonic Cyclic Symmetry Analysis in the
        Cyclic Symmetry Analysis Guide.
        """
        command = f"RCYC,{ir},{ia},{sector},{name}"
        return self.run(command, **kwargs)

    def resp(
        self,
        ir="",
        lftab="",
        ldtab="",
        spectype="",
        dampratio="",
        dtime="",
        tmin="",
        tmax="",
        inputtype="",
        **kwargs,
    ):
        """Generates a response spectrum.

        APDL Command: RESP

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the response spectrum
            results (2 to NV [NUMVAR]).  If this number is the same as for a
            previously defined variable, the previously defined variable will
            be overwritten with these results.

        lftab
            Reference number of variable containing frequency table (created
            with FILLDATA or DATA command).  The frequency table defines the
            number and frequency of oscillating systems used to determine the
            response spectrum. The frequency interval need not be constant over
            the entire range. Frequencies must be input in ascending order.

        ldtab
            Reference number of variable containing the input time-history.

        spectype
            Defines the type of response spectrum to be calculated:

            0 or 1 - Displacement (relative to base excitation)

            2 - Velocity (relative to base excitation)

            3 - Acceleration response spectrum (absolute)

            4 - Pseudo-velocity

            5 - Pseudo-acceleration

        dampratio
            Ratio of viscous damping to critical damping (input as a decimal
            number).

        dtime
            Integration time step. This value should be equal to or greater
            than the integration time step used in the initial transient
            analysis performed to generate the input time-history (LDTAB).

        tmin, tmax
            Specifies a subset of the displacement-time history to be used in
            the response spectrum calculation.  Defaults to the full time
            range.

        inputtype
            Defines the type of the input time-history:

            0 - Displacement (default)

            1 - Acceleration

        Notes
        -----
        This command generates a response spectrum from a displacement or
        acceleration time-history and frequency data. The response spectrum is
        defined as the maximum response of single degree of freedom systems of
        varying frequency (or period) to a given input support excitation.

        A response spectrum analysis (ANTYPE, SPECTR with SPOPT, SPRS or MPRS)
        requires a response spectrum input. This input can be determined from
        the response spectrum printout or display of this command.

        If a response spectrum is to be calculated from a given displacement
        (or acceleration) time-history, the displacement time-history may be
        input to a single one-element reduced linear transient dynamic
        (ANTYPE,TRANS) analysis, so that the calculated output (which should be
        the same as the input) will be properly located on the file.

        The integration time step (argument DTIME on the RESP command) and the
        damping coefficient (argument dampRatio) are constant over the
        frequency range. The number of calculations done per response spectrum
        curve is the product of the number of input solution points (TMAX-
        TMIN)/DTIME and the number of frequency points (frequencies located in
        variable LFTAB).

        Input solution points requested (using DTIME and the frequency range)
        at a time not corresponding to an actual displacement solution time on
        the file are linearly interpolated with respect to the existing points.

        For the details of the response spectrum calculation, see POST26 -
        Response Spectrum Generator (RESP).
        """
        command = f"RESP,{ir},{lftab},{ldtab},{spectype},{dampratio},{dtime},{tmin},{tmax},{inputtype}"
        return self.run(command, **kwargs)

    def rpsd(
        self, ir="", ia="", ib="", itype="", datum="", name="", signif="", **kwargs
    ):
        """Calculates response power spectral density (PSD).

        APDL Command: RPSD

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previous
            variable, the previous variable will be overwritten with this
            result.

        ia, ib
            Reference numbers of the two variables to be operated on.  If only
            one, leave IB blank.

        itype
            Defines the type of response PSD to be calculated:

            0,1 - Displacement (default).

            2 - Velocity.

            3 - Acceleration.

        datum
            Defines the reference with respect to which response PSD is to be
            calculated:

            1 - Absolute value.

            2 - Relative to base (default).

        name
            Thirty-two character name identifying variable on listings and
            displays.  Embedded blanks are compressed for output.

        signif
            Combine only those modes whose significance level exceeds the
            SIGNIF threshold. The significance level is defined as the modal
            covariance matrix term divided by the maximum of all the modal
            covariance matrix terms. Any term whose significance level is less
            than SIGNIF is considered insignificant and does not contribute to
            the response. All modes are taken into account by default (SIGNIF =
            0.0).

        Notes
        -----
        This command calculates response power spectral density (PSD) for the
        variables referenced by the reference numbers IA and IB.  The variable
        referred by IR will contain the response PSD.  You must issue the
        STORE,PSD command first; File.PSD must be available for the
        calculations to occur.

        See POST26 - Response Power Spectral Density in the Mechanical APDL
        Theory Reference for more information on these equations.
        """
        command = f"RPSD,{ir},{ia},{ib},{itype},{datum},{name},{signif}"
        return self.run(command, **kwargs)

    def smooth(
        self,
        vect1="",
        vect2="",
        datap="",
        fitpt="",
        vect3="",
        vect4="",
        disp="",
        **kwargs,
    ):
        """Allows smoothing of noisy data and provides a graphical representation

        APDL Command: SMOOTH
        of the data.

        Parameters
        ----------
        vect1
            Name of the first vector that contains the noisy data set (i.e.,
            independent variable). You must create and fill this vector before
            issuing SMOOTH.

        vect2
            Name of the second vector that contains the dependent set of data.
            Must be the same length as the first vector. You must create and
            fill this vector before issuing SMOOTH.

        datap
            Number of data points to be fitted, starting from the beginning of
            the vector. If left blank, the entire vector will be fitted. The
            maximum number of data points is 100,000 (or greater, depending on
            the memory of the computer).

        fitpt
            Order of the fitting curve that will be used as a smooth
            representation of the data. This number should be less than or
            equal to the number of the data points. Default (blank) is one-half
            the number of data points. Maximum number of smoothed data fitting
            order is the number of data points up to 50.  Depending on this
            number, the smoothed curve will be one of the following:

            1 - Curve is the absolute average of all of the data points.

            2 - Curve is the least square average of all of the data points.

            3 or more - Curve is a polynomial of the order (n-1), where n is the number of data fitting
                        order points.

        vect3
            Name of the vector that contains the smoothed data of the
            independent variable. This vector should have a length equal to or
            greater than the number of smoothed data points. In batch (command)
            mode, you must create this vector before issuing the SMOOTH
            command. In interactive mode, the GUI automatically creates this
            vector (if it does not exist). If you do not specify a vector name,
            the GUI will name the vector smth_ind.

        vect4
            Name of the vector that contains the smoothed data of the dependent
            variable.  This vector must be the same length as Vect3.  In batch
            (command) mode, you must create this vector before issuing the
            SMOOTH command. In interactive mode, the GUI automatically creates
            this vector (if it does not exist). If you do not specify a vector
            name, the GUI will name the vector smth_dep.

        disp
            Specifies how you want to display data. No default; you must
            specify an option.

            1 - Unsmoothed data only

            2 - Smoothed data only

            3 - Both smoothed and unsmoothed data

        Notes
        -----
        You can control the attributes of the graph using standard ANSYS
        controls (/GRID, /GTHK, /COLOR, etc.). If working interactively, these
        controls appear in this dialog box for convenience, as well as in their
        standard dialog boxes. You must always create Vect1 and Vect2 (using
        ``*DIM``) and fill these vectors before smoothing the data. If you're
        working interactively, ANSYS automatically creates Vect3 and Vect4, but
        if you're working in batch (command) mode, you must create Vect3 and
        Vect4 (using ``*DIM``) before issuing SMOOTH.  Vect3 and Vect4 are then
        filled automatically by ANSYS.  In addition, ANSYS creates an
        additional TABLE type array that contains the smoothed array and the
        unsmoothed data to allow for plotting later with ``*VPLOT``.  Column 1 in
        this table corresponds to Vect1, column 2 to Vect2, and column 3 to
        Vect4.  This array is named Vect3_SMOOTH, up to a limit of 32
        characters. For example, if the array name is X1, the table name is
        X1_SMOOTH.

        This command is also valid in PREP7 and SOLUTION.
        """
        command = f"SMOOTH,{vect1},{vect2},{datap},{fitpt},{vect3},{vect4},{disp}"
        return self.run(command, **kwargs)

    def vget(self, par="", ir="", tstrt="", kcplx="", **kwargs):
        """Moves a variable into an array parameter vector.

        APDL Command: VGET

        Parameters
        ----------
        par
            Array parameter vector in the operation.

        ir
            Reference number of the variable (1 to NV [NUMVAR]).

        tstrt
            Time (or frequency) corresponding to start of IR data.  If between
            values, the nearer value is used.

        kcplx
            Complex number key:

            * ``0`` - Use the real part of the IR data.

            * ``1`` - Use the imaginary part of the IR data.

        Notes
        -----
        Moves a variable into an array parameter vector.  The starting array
        element number must be defined.  For example, VGET,A(1),2 moves
        variable 2 (starting at time 0.0) to array parameter A.  Looping
        continues from array element A(1) with the index number incremented by
        one until the variable is filled.  The number of loops may be
        controlled with the ``*VLEN`` command (except that loop skipping (NINC) is
        not allowed).  For multi-dimensioned array parameters, only the first
        (row) subscript is incremented.
        """
        command = f"VGET,{par},{ir},{tstrt},{kcplx}"
        return self.run(command, **kwargs)

    def vput(self, par="", ir="", tstrt="", kcplx="", name="", **kwargs):
        """Moves an array parameter vector into a variable.

        APDL Command: VPUT

        Parameters
        ----------
        par
            Array parameter vector in the operation.

        ir
            Arbitrary reference number assigned to this variable (1 to NV
            [NUMVAR]).  Overwrites any existing results for this variable.

        tstrt
            Time (or frequency) corresponding to start of IR data.  If between
            values, the nearer value is used.

        kcplx
            Complex number key:

            0 - Use the real part of the IR data.

            1 - Use the imaginary part of the IR data.

        name
            Thirty-two character name identifying the item on printouts and
            displays. Defaults to the label formed by concatenating VPUT with
            the reference number IR.

        Notes
        -----
        At least one variable should be defined (NSOL, ESOL, RFORCE, etc.)
        before using this command.  The starting array element number must be
        defined.  For example, VPUT,A(1),2 moves array parameter A to variable
        2 starting at time 0.0.  Looping continues from array element A(1) with
        the index number incremented by one until the variable is filled.
        Unfilled variable locations are assigned a zero value.  The number of
        loops may be controlled with the ``*VLEN`` command (except that loop
        skipping (NINC) is not allowed).  For multi-dimensioned array
        parameters, only the first (row) subscript is incremented.
        """
        command = f"VPUT,{par},{ir},{tstrt},{kcplx},{name}"
        return self.run(command, **kwargs)



from typing import Union

MapdlInt = Union[
    str, int
]  # mapdl integers can either be strings or integers. "1" and 1 are equally valid
MapdlFloat = Union[
    str, float
]  # mapdl floats can either be strings or floats. "1.0" and 1.0 are equally valid



"""These SESSION commands are for file operations, such as deleting,
copying, and listing."""


class Files:
    def anstoasas(self, fname="", key="", **kwargs):
        """Creates an ASAS input file from the current ANSYS model.

        APDL Command: ANSTOASAS

        Parameters
        ----------
        fname
            ASAS file name. Defaults to Jobname.

        key
            Key indicating type of file to produce:

            0 - ASAS file for use by ANSYS Aqwa (no loads written). Creates the file
                Fname.asas.

            1 - ASAS file (all data written, including loads). Creates the file Fname.asas.

            2 - ASAS(NL) file. Creates the file Fname.asnl.

        Notes
        -----
        This command creates an input file for the ANSYS Asas Finite Element
        Analysis System from the model and loads currently in the database,
        based on the currently selected set of elements. Most common structural
        element types are written, as well as sections (or real constants),
        materials, boundary conditions and loads, and solution and load step
        options.

        Data Written

        The following data is written:

        Solution control options

        Nodes

        Elements

        Material data

        Geometry data

        Section data

        ANSYS element components (ASAS sets)

        Boundary conditions

        Loads

        Added mass (via MASS21 element)

        Details are provided in the following sections.

        Not all data is written. You must verify the completeness and accuracy
        of the data. Only loading at the current step is transferred; hence, no
        load step history is captured.

        Solution Control Options

        The ASAS project name is defined as "ANSYS".

        The solution control options are converted as follows:

        JOB: STAT SPIT: KGEOM

        JOB: STAT SPIT: KGEOM

        For conversion to ASAS(NL), the large displacement option is set based
        on NLGEOM, final load solution time is set based on TIME, and sub-step
        times are set based on DELTIM or NSUBST (assuming constant step size).

        Element Data

        If you intend to use the data only with AQWA-WAVE, only the elements
        that form the wetted surface are required.  Selecting these elements
        before invoking the ANSTOASAS command will improve performance.  In
        order for AQWA-WAVE to identify the direction of the wave loading, all
        elements must be defined by nodes in a clockwise direction. For further
        information, refer to the AQWA-WAVE manual.

        The element types are converted as follows:

        SPR1:  SPR2: if: rotational: spring:  FLA2: (ASAS(L): only):  if:
        nodes: are: not: coincident:  and: longitudinal: spring:

        QUM4:  TRM3: -: if: Triangular

        BRK8:  TET4: -: if: Tetrahedral:  BRK6: -: if: Prism

        QUS4:  TBC3: -: if: Triangular

        QUM8:  TRM6: -: if: Triangular

        BR20:  TE10: -: if: Tetrahedral: :  BR15: -: if: Prism

        QUS4:  TBC3: -: if: Triangular

        QUM4:  TRM3: -: if: Triangular

        QUM8:  TRM6: -: if: Triangular

        BRK8:  TET4: -: if: Tetrahedral:  BRK6: -: if: Prism

        BR20:  TE10: -: if: Tetrahedral:  BR15: -: if: Prism

        TCBM: -: if: ASAS(L):  STF4: -: if: ASAS(NL)

        Documentation for this legacy element type appears in the Feature
        Archive.

        Material Data

        Linear isotropic material conversion is supported for ASAS and
        ASAS(NL).

        Geometry Data
        """
        command = "ANSTOASAS,%s,%s" % (str(fname), str(key))
        return self.run(command, **kwargs)

    def anstoaqwa(
        self,
        fname="",
        vertaxis="",
        gc="",
        rho="",
        hwl="",
        diffkey="",
        symxkey="",
        symykey="",
        **kwargs,
    ):
        """Creates an AQWA-LINE input file from the current ANSYS model.

        APDL Command: ANSTOAQWA

        Parameters
        ----------
        fname
            AQWA file name. Defaults to Jobname.

        vertaxis
            Axis in the vertical direction:

            Y (or 2)  - Global Y axis.

            Z (or 3)  - Global Z axis (default).

        gc
            Gravitational acceleration. Defaults to 9.81.

        rho
            Density of water. Defaults to 1025.0.

        hwl
            Waterline height in model coordinates. Defaults to 0.0.

        diffkey
            Diffracting model key:

            0 - Create a non-diffracting AQWA model.

            1 - Create a diffracting AQWA model (default).

        symxkey
            Key indicating if model is symmetric about the global XZ plane:

            0 - No symmetry about XZ plane (default).

            1 - Use symmetry about XZ plane. Only include (or select) half the model.

        symykey
            Key indicating if model is symmetric about the global YZ plane:

            0 - No symmetry about YZ plane (default).

            1 - Use symmetry about YZ plane. Only include (or select) half the model.

        Notes
        -----
        This command creates the input file Fname.aqwa for the ANSYS Aqwa
        Multi-Body Hydrodynamics System for diffraction analysis in AQWA-LINE
        from the model currently in the database, based on the currently
        selected set of elements. The selected set must only include the hull
        envelope; no internal structure should be selected.

        There should be a line of nodes defined at the waterline. Only those
        elements that are entirely below the waterline will be specified as
        diffracting. If there are no waterline nodes, there will be no
        diffracting elements at the waterline, which will severely reduce the
        accuracy of the diffraction analysis.

        The translator maps PLANE42, SHELL41, SHELL63, and SHELL181 elements to
        PANELs, and maps PIPE16 and PIPE59 elements to TUBEs. It does not
        recognize any other element types. Any material or geometric properties
        can be used for the shell elements, as AQWA does not need any
        properties at all and the command does not use them. All the shell
        elements below the water must have their normals pointing outward.

        TUBE elements in AQWA have material density, outside diameter, wall
        thickness, added mass, and drag coefficients, so appropriate properties
        should be used in the ANSYS model. PIPE59 elements can have added mass
        and damping coefficients; these will be written to the file. The ANSYS
        program uses the inertia coefficient CM, whereas AQWA uses the added
        mass coefficient CA, where CM = (1 + CA). This correction is made
        automatically.

        In AQWA the vertical axis is always the Z-axis. The command can convert
        a model built with either the Y or Z-axis vertical, but the X-axis must
        be horizontal and should preferably be along the fore/aft axis of the
        vessel.  If the structure is symmetric and you wish to use the symmetry
        options, you must only select one half or one quarter of the model, as
        appropriate. If you model a complete vessel and specify X symmetry, the
        AQWA model will contain two sets of coincident elements.

        If you are working from a model created for a structural analysis, it
        will probably be necessary to remesh the model as the structural mesh
        is most likely finer than needed for a diffraction analysis.

        If you enter this command interactively (with the GUI active) and no
        data is provided for the command options, you will be prompted for
        their values.

        You must verify the completeness and accuracy of the data written.
        """
        command = "ANSTOAQWA,%s,%s,%s,%s,%s,%s,%s,%s" % (
            str(fname),
            str(vertaxis),
            str(gc),
            str(rho),
            str(hwl),
            str(diffkey),
            str(symxkey),
            str(symykey),
        )
        return self.run(command, **kwargs)

    def assign(self, ident="", fname="", ext="", lgkey="", **kwargs):
        """Reassigns a file name to an ANSYS file identifier.

        APDL Command: /ASSIGN

        Parameters
        ----------
        ident
            ANSYS file name identifier.  Valid identifiers are: CMS, EMAT,
            EROT,  ESAV, FULL, LN07, LN09, LN11, LN20, LN21, LN22, LN25, LN31,
            LN32, MODE, OSAV, RDSP, RFRQ, RMG, RST, RSTP, RTH, SELD, and SSCR.
            See File Management and Files for file descriptions.  If blank,
            list currently reassigned files.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        lgkey
            Key to specify local or global file name control for the specified
            file identifier in a distributed-memory parallel processing
            (Distributed ANSYS) run. For more information on local and global
            files, see File Handling Conventions in the Parallel Processing
            Guide.

            BOTH - Reassign the file name for both the local and global files (default).

            LOCAL - Reassign the file name for only the local files.

            GLOBAL - Reassign the file name for only the global file.

        Notes
        -----
        The reassignment of file names is valid only if it is done before the
        file is used.  All file reassignments are retained (not cleared) even
        if the database is cleared [/CLEAR] or the Jobname is changed
        [/FILNAME].  Assigned files may be overwritten.  If file name arguments
        (Fname, Ext, --) are blank, the default ANSYS assignment is restored.
        Use SEOPT for SUB files and SEEXP for DSUB files.

        This command is valid only at the Begin Level.

        This command also checks to ensure that the path/file is valid and can
        be written by the user. If it is not valid, an error message will be
        returned. Ensure that the directory exists prior to using /ASSIGN
        command.
        """
        return self.run(f"/ASSIGN,{ident},{fname},{ext},,{lgkey}", **kwargs)

    def slashclog(self, fname="", ext="", **kwargs):
        """APDL Command: /CLOG

        Copies the session log file to a named file.

        Parameters
        ----------
        fname
            File name and directory path to which the log file is to be copied
            (248 characters maximum, including directory). If you do not
            specify a directory path, it will default to your working directory
            and you can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        This command is valid in any processor, but only during an interactive
        run.
        """
        command = "/CLOG,%s,%s" % (str(fname), str(ext))
        return self.run(command, **kwargs)

    def copy(self, fname1="", ext1="", fname2="", ext2="", distkey="", **kwargs):
        """Copies a file.

        APDL Command: /COPY

        Parameters
        ----------
        fname1
            File name to be copied and its directory path (248
            characters maximum for both file name and directory). If
            you do not specify a directory path, it will default to
            your working directory and you can use all 248 characters
            for the file name.

        ext1
            Filename extension (eight-character maximum).

        fname2
            File name to be created and its directory path (248
            characters maximum for both file name and directory). If
            you do not specify a directory path, it will default to
            your working directory and you can use all 248 characters
            for the file name.

        ext2
            Filename extension (eight-character maximum).

        distkey
            Key that specifies whether the copy operation is performed
            on all processes in distributed parallel mode (Distributed
            ANSYS):

            0 (OFF or NO) - The program performs the copy operation
            only on the master process (default).

            1 (ON or YES) - The program performs the copy operation
            locally on each process.

            2 or BOTH - The program performs the copy operation for
            Fname.Ext on the master process and for FnameN.Ext on all
            processes.

        Notes
        -----
        The original file is untouched.  Ex: /COPY,A,,,B copies file A
        to B in the same directory.  /COPY,A,DAT,,,INP copies the file
        A.DAT to A.INP.  See the Operations Guide for details.  ANSYS
        binary and ASCII files can be copied.

        In distributed parallel mode (Distributed ANSYS), only the
        master process will copy Fname1.Ext1 to Fname2.Ext2 by
        default. However, when DistKey is set to 1 (or ON or YES), the
        command is executed by all processes. In this case, Fname1 and
        Fname2 will automatically have the process rank appended to
        them. This means Fname1N.Ext1 will be copied to Fname2N.Ext2
        by all processes, where N is the Distributed ANSYS process
        rank.  For more information see Differences in General
        Behavior in the Parallel Processing Guide.
        """
        command = f"/COPY,{fname1},{ext1},,{fname2},{ext2},,{distkey}"
        return self.run(command, **kwargs)

    def fcomp(self, ident="", level="", **kwargs):
        """Specifies file compression level.

        Parameters
        ----------
        ident
            ANSYS file name identifier. Input the label RST to compress
            the following results files: .RST, .RSTP, .RTH, and .RMG. See
            File Management and Files for file descriptions.

        level
            Compression level. Valid input values are 0 (no compression -
            default) to 5 (maximum compression).

        Notes
        -----
        Command Default
        File compression is not performed.

        Specifies file compression for results files (.RST, .RSTP, .RTH,
        and .RMG files). Records are compressed as they are written and
        uncompressed as they are read (for example, by the SET command).

        See File Compression in the Basic Analysis Guide for more details.
        """
        return self.run(f"/FCOMP,{ident},{level}", **kwargs)

    def lgwrite(self, fname="", ext="", kedit="", **kwargs):
        """Writes the database command log to a file.

        APDL Command: LGWRITE

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

            The file name defaults to Jobname.

        ext
            Filename extension (eight-character maximum).

        kedit
            Flag to suppress nonessential commands:

            NONE - Do not suppress any commands (default).

            COMMENT - Write nonessential commands as comments (starting with !).

            REMOVE - Do not write nonessential commands or comments.

        Notes
        -----
        Writes the database command log to a named file.  The database
        command log contains all commands that were used to create the
        current database.  These commands are recorded in the database
        as they are issued, and saved in the database file (File.DB)
        whenever the database is saved.  The LGWRITE command extracts
        these commands from the database and writes them to a file.
        Nonessential commands (for listing, graphics displays, help,
        etc.) can be excluded from the file by using the Kedit field.
        The file resulting from LGWRITE can be used as command input
        to the program.  This command is most useful if the session
        log file (File.LOG), which is normally saved during an
        interactive session, has been lost or corrupted.

        This command is valid in any processor.
        """
        return self.run(f"LGWRITE,{fname},{ext},,{kedit}", **kwargs)

    def starlist(self, fname="", ext="", **kwargs):
        """Displays the contents of an external, coded file.

        APDL Command: ``*LIST``

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).  An
            unspecified directory path defaults to the working directory;
            in this case, you can use all 248 characters for the file
            name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        Displays the contents of an external, coded file.  The file to be
        listed cannot be in use (open) at the time (except for the error
        file, File.ERR, which may be displayed with ``*LIST,ERR``).

        This command is valid in any processor.
        """
        command = f"*LIST,{fname},{ext}"
        return self.run(command, **kwargs)

    def rename(self, fname1="", ext1="", fname2="", ext2="", distkey="", **kwargs):
        """Renames a file.

        APDL Command: /RENAME

        Parameters
        ----------
        fname1
            The file to be renamed. You can also include an optional
            directory path as part of the specified file name; if not,
            the default file location is the working directory.

        ext1
            Filename extension (eight-character maximum).

        fname2
            The new name for the file. You can also include an
            optional directory path as part of the new file name; if
            not, the default is the working directory. A maximum of
            248 characters is allowed for the file name (or combined
            file name and directory path, if both are specified).

        ext2
            Filename extension (eight-character maximum).

        distkey
            Key that specifies whether the rename operation is
            performed on all processes in distributed parallel mode
            (Distributed ANSYS):

            1 (ON or YES) - The program performs the rename operation
            locally on each process.

            0 (OFF or NO) - The program performs the rename operation
            only on the master process (default).

        Notes
        -----
        Renames a file.  Ex: /RENAME,A,,,B renames file A to B in the
        same directory.  /RENAME,A,DAT,,,INP renames file A.DAT to
        A.INP. On all systems, this command will overwrite any
        existing file named B. See the Operations Guide for
        details. Only ANSYS binary files should be renamed. Use /SYS
        and system renaming commands for other files.

        In distributed parallel mode (Distributed ANSYS), only the
        master process will rename Fname1.Ext1 to Fname2.Ext2 by
        default. However, when DistKey is set to 1 (or ON or YES), the
        command is executed by all processes. In this case, Fname1 and
        Fname2 will automatically have the process rank appended to
        them. This means Fname1N.Ext1 will be renamed to Fname2N.Ext2
        by all processes, where N is the Distributed ANSYS process
        rank. For more information see Differences in General Behavior
        in the Parallel Processing Guide.

        Renaming across system partitions may be internally done by a
        copy and delete operation on some systems.

        This command is valid only at the Begin Level.
        """
        return self.run(
            f"/RENAME,{fname1},{ext1},,{fname2},{ext2},,{distkey}", **kwargs
        )

    def slashfdele(self, ident="", stat="", **kwargs):
        """Deletes a binary file after it is used.

        APDL Command: /FDELE

        Parameters
        ----------
        ident
            ANSYS file name identifier.  Valid identifiers are:  EMAT, ESAV,
            FULL, SUB, MODE, DSUB, USUB, OSAV, and SELD.  See the Basic
            Analysis Guide for file descriptions.

        stat
            Keep or delete key:

            KEEP - Keep this file.

            DELE - Delete (or do not write, if not necessary) this file.

        Notes
        -----
        Deletes as soon as possible (or prevents writing) a binary file created
        by the ANSYS program to save space.

        .. warning::
            Deleting files that are necessary for the next substep,
            load step, or analysis will prevent continuation of the
            run.

        This command is valid only at the Begin Level.
        """
        command = f"/FDELE,{ident},{stat}"
        return self.run(command, **kwargs)

    def slashdelete(self, fname="", ext="", distkey="", **kwargs):
        """Deletes a file.

        APDL Command: /DELETE

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        distkey
            Key that specifies whether the file deletion is performed
            on all processes in distributed parallel mode (Distributed
            ANSYS):

            1 (ON or YES) - The program performs the file deletion
            locally on each process.

            0 (OFF or NO) - The program performs the file deletion
            only on the master process (default).

        Notes
        -----
        In distributed parallel mode (Distributed ANSYS), only the
        master process will delete Fname.Ext by default. However, when
        DistKey is set to 1 (or ON or YES), the command is executed by
        all processes. In this case, Fname will automatically have the
        process rank appended to it.  This means FnameN.Ext will be
        deleted by all processes, where N is the Distributed ANSYS
        process rank. For more information see Differences in General
        Behavior in the Parallel Processing Guide.
        """
        return self.run(f"/DELETE,{fname},{ext},,{distkey}", **kwargs)
